/*
 * Softcam plugin to VDR (C++)
 *
 * This code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/time.h>

#include <linux/dvb/ca.h>
#include <vdr/ci.h>
#include <vdr/dvbdevice.h>
#include <vdr/dvbci.h>
#include <vdr/channels.h>
#include <vdr/thread.h>

#include "FFdecsa/FFdecsa.h"

#include "common.h"
#include "cam.h"
#include "sc.h"
#include "scsetup.h"
#include "filter.h"
#include "system.h"
#include "data.h"
#include "misc.h"

#define IDLE_SLEEP          0 // idleTime when sleeping
#define IDLE_GETCA        200 // idleTime when waiting for ca descriptors
#define IDLE_GETCA_SLOW 20000 // idleTime if no enc. system
#define IDLE_NO_SYNC      800 // idleTime when not in sync
#define IDLE_SYNC        2000 // idleTime when in sync

#define CW_REPEAT_TIME   2000 // rewrite CW after X ms
#define LOG_COUNT           3 // stop logging after X complete ECM cycles
#define CHAIN_HOLD     120000 // min. time to hold a logger chain
#define ECM_DATA_TIME    6000 // time to wait for ECM data updates
#define ECM_UPD_TIME   120000 // delay between ECM data updates
#define MAX_ECM_IDLE   300000 // delay before an idle handler can be removed
#define MAX_ECM_HOLD    15000 // delay before an idle handler stops processing

#define DEBUG_PIDS
//#define DEBUG_AV7110
#define DEBUG_LOG
//#define DEBUG_LOG_EXTRA
#define DEBUG_LOG_STATS
//#define DEBUG_CSA
#define DEBUG_CI

// Hex-Data Logging
// LOG_DUMP to enable, rest to select output
//#define LOG_DUMP "/var/tmp/hexdump.txt"
#ifdef LOG_DUMP
//#define LOG_ECM
//#define LOG_EMM
//#define LOG_CAT
//#define LOG_PMT
#endif

#ifdef DEBUG_PIDS
#define dp(x) { x; }
#else
#define dp(x) ;
#endif

#ifdef DEBUG_LOG
#define dl(x) { (x); }
#else
#define dl(x) ;
#endif

#ifdef DEBUG_LOG_EXTRA
#define dle(x) { (x); }
#else
#define dle(x) ;
#endif

#ifdef DEBUG_CSA
#define dc(x) { x; }
#else
#define dc(x) ;
#endif

#ifdef DEBUG_CI
#define di(x) { x; }
#define DUMP(data,len) do { const int l=(len); \
                         for(int i=0 ; i<l ; i++) printf(" %02x",*((data)+i)); \
                         printf("\n"); } while(0);
#else
#define di(x) ;
#define DUMP(data,len)
#endif

#if APIVERSNUM < 10338
#define CA_ENCRYPTED_MIN CACONFBASE
#endif

// -----------------------------------------------------------------------------

#ifdef DEBUG_PIDS
static const char *typeNames[] = { "typ0","typ1","VIDEO","typ3","AUDIO","typ5","DOLBY","typ6+" };
#define TYPENAME(type) (typeNames[(type)<=7?(type):7])
#endif

#ifdef LOG_DUMP
#define PER_LINE 16
#include <stdarg.h>
void LogDump(const unsigned char *buffer, int n, const char *format, ...) __attribute__ ((format (printf,3,4)));
void LogDump(const unsigned char *buffer, int n, const char *format, ...)
{
  static cMutex logMutex;
  static FILE *f=0;

  time_t t=time(0);
  logMutex.Lock();
  if(!f) f=fopen(LOG_DUMP,"a");
  if(f) {
    struct tm tm_r;
    struct tm *tm=localtime_r(&t,&tm_r);
    char timestr[32];
    strftime(timestr,sizeof(timestr),"%b %d %T",tm);
    fprintf(f,"%s (size=%d/0x%04x) ",timestr,n,n);
    va_list ap; va_start(ap,format); vfprintf(f,format,ap); va_end(ap);
    fprintf(f,"\n");
    for(int i=0 ; i<n ;) {
      fprintf(f,"%04x:",i);
      for(int l=0 ; l<PER_LINE && i<n ; l++) fprintf(f," %02x",buffer[i++]);
      fprintf(f,"\n");
      }
    fflush(f);
    }
  logMutex.Unlock();
}
#endif

// -- cLogStats ---------------------------------------------------------------

#ifdef DEBUG_LOG_STATS

#define COUNTS  20
#define SAMPLE (30*1000)
#define AVR1   (60*1000)
#define AVR2   (4*60*1000)
#define AVR3   (10*60*1000)
#define REPORT (60*1000)

class cLogStats : public cThread {
private:
  cTimeMs sTime, repTime;
  int sCount, sIdx, sCounts[COUNTS][2];
protected:
  virtual void Action(void);
public:
  cLogStats(void);
  ~cLogStats();
  void Count(void);
  };

static cMutex logstatsMutex;
static cLogStats *logstats=0;

void LogStatsUp(void)
{
  logstatsMutex.Lock();
  if(!logstats) logstats=new cLogStats;
  logstatsMutex.Unlock();
}

void LogStatsDown(void)
{
  logstatsMutex.Lock();
  if(logstats) { delete logstats; logstats=0; }
  logstatsMutex.Unlock();
}

cLogStats::cLogStats(void)
{
  sCount=sIdx=0;
  for(int i=0; i<COUNTS; i++) { sCounts[i][0]=0; sCounts[i][1]=SAMPLE; }
  SetDescription("logger stats");
  Start();
}

cLogStats::~cLogStats()
{
  Cancel(2);
}

void cLogStats::Count(void)
{
  sCount++;
}

void cLogStats::Action(void)
{
  while(Running()) {
    cCondWait::SleepMs(50);
    if(sTime.Elapsed()>SAMPLE) {
      sCounts[sIdx][0]=sCount;          sCount=0;
      sCounts[sIdx][1]=sTime.Elapsed(); sTime.Set();
      if(++sIdx >= COUNTS) sIdx=0;
      }
    if(repTime.Elapsed()>REPORT) {
      repTime.Set();
      if(sCounts[(sIdx+COUNTS-1)%COUNTS][0]>0) {
        printf("logstats: EMM packet load average (%d/%d/%dmin)",AVR1/60000,AVR2/60000,AVR3/60000);
        int s=0, t=0;
        for(int i=1; i<=COUNTS; i++) {
          s+=sCounts[(sIdx+COUNTS-i)%COUNTS][0];
          t+=sCounts[(sIdx+COUNTS-i)%COUNTS][1];
          if(i==(AVR1/SAMPLE) || i==(AVR2/SAMPLE) || i==(AVR3/SAMPLE))
            printf(" %4d",(int)((float)s/(float)t*1000.0));
          }
        printf(" pks/s\n");
        }
      }
    }
}

#endif

// -- cLogChain ----------------------------------------------------------------

class cLogChain : public cSimpleItem {
public:
  int cardNum, caid;
  bool softCSA, active, delayed;
  cTimeMs delay;
  cPids pids;
  cSimpleList<cSystem> systems;
  //
  cLogChain(int CardNum, bool soft);
  void Process(int pid, unsigned char *data);
  bool Parse(const unsigned char *cat);
  };

cLogChain::cLogChain(int CardNum, bool soft)
{
  cardNum=CardNum; softCSA=soft; active=delayed=false;
}

void cLogChain::Process(int pid, unsigned char *data)
{
  if(active) {
    for(cSystem *sys=systems.First(); sys; sys=systems.Next(sys))
      sys->ProcessEMM(pid,caid,data);
    }
}

bool cLogChain::Parse(const unsigned char *cat)
{
  if(cat[0]==0x09) {
    caid=WORD(cat,2,0xFFFF);
    dl(printf("logger %d: chain caid %04x",cardNum,caid))
    cSystem *sys;
    if(systems.Count()>0) {
      dl(printf(" ++"))
      for(sys=systems.First(); sys; sys=systems.Next(sys))
        sys->ParseCAT(&pids,cat);
      }
    else {
      dl(printf(" ->"))
      int Pri=0;
      while((sys=cSystems::FindBySysId(caid,!softCSA,Pri))) {
        Pri=sys->Pri();
        if(sys->HasLogger()) {
          sys->CardNum(cardNum);
          sys->ParseCAT(&pids,cat);
          systems.Add(sys);
          dl(printf(" %s(%d)",sys->Name(),sys->Pri()))
          }
        else
          delete sys;
        }
      }
#ifdef DEBUG_LOG
    if(systems.Count()==0) printf(" none available");
    for(cPid *pid=pids.First(); pid; pid=pids.Next(pid))
       printf(" [%04x-%02x/%02x/%02x]",pid->pid,pid->sct,pid->mask,pid->mode);
    printf("\n");
#endif
    if(systems.Count()>0 && pids.Count()>0)
      return true;
    }
  return false;
}

// -- cLogger ------------------------------------------------------------------

class cLogger : public cAction {
private:
  int cardNum;
  bool softCSA, up;
  cSimpleList<cLogChain> chains;
  cSimpleList<cEcmInfo> active;
  //
  cPidFilter *catfilt;
  int catVers;
  //
  cPidFilter *AddFilter(int Pid, int Section, int Mask, int Mode, int IdleTime, bool Crc);
  void SetChains(void);
  void ClearChains(void);
  void StartChain(cLogChain *chain);
  void StopChain(cLogChain *chain, bool force);
protected:
  virtual void Process(cPidFilter *filter, unsigned char *data, int len);
public:
  cLogger(int CardNum, bool soft);
  virtual ~cLogger();
  void EcmStatus(const cEcmInfo *ecm, bool on);
  void Up(void);
  void Down(void);
  };

cLogger::cLogger(int CardNum, bool soft)
:cAction("logger",CardNum)
{
  cardNum=CardNum; softCSA=soft;
  catfilt=0; up=false;
  Priority(10);
}

cLogger::~cLogger()
{
  Down();
}

void cLogger::Up(void)
{
  Lock();
  if(!up) {
    dle(printf("logger %d: UP\n",cardNum))
    catVers=-1;
    catfilt=AddFilter(1,0x01,0xFF,0,0,true);
    up=true;
    }
  Unlock();
}

void cLogger::Down(void)
{
  Lock();
  if(up) {
    dle(printf("logger %d: DOWN\n",cardNum))
    ClearChains();
    DelAllFilter();
    catfilt=0; up=false;
    }
  Unlock();
}

void cLogger::EcmStatus(const cEcmInfo *ecm, bool on)
{
  Lock();
  dle(printf("logger %d: ecm prgid=%d caid=%04x prov=%.4x %s\n",cardNum,ecm->prgId,ecm->caId,ecm->provId,on ? "active":"inactive"))
  cEcmInfo *e;
  if(on) {
    e=new cEcmInfo(ecm);
    active.Add(e);
    if(!up) Up();
    }
  else {
    for(e=active.First(); e; e=active.Next(e))
      if(e->Compare(ecm)) {
        active.Del(e);
        break;
        }
    }
  SetChains();
  Unlock();
}

void cLogger::SetChains(void)
{
  for(cLogChain *chain=chains.First(); chain; chain=chains.Next(chain)) {
    bool act=false;
    if(ScSetup.AutoUpdate>1) act=true;
    else if(ScSetup.AutoUpdate==1) {
      for(cEcmInfo *e=active.First(); e; e=active.Next(e))
        if((e->emmCaId && chain->caid==e->emmCaId) || chain->caid==e->caId) {
          act=true; break;
          }
       }
    if(act) StartChain(chain);
    else StopChain(chain,false);
    }
}

void cLogger::ClearChains(void)
{
  for(cLogChain *chain=chains.First(); chain; chain=chains.Next(chain))
    StopChain(chain,true);
  chains.Clear();
}

void cLogger::StartChain(cLogChain *chain)
{
  if(chain->delayed)
    dle(printf("logger %d: restarting delayed chain %04x\n",cardNum,chain->caid))
  chain->delayed=false;
  if(!chain->active) {
    dl(printf("logger %d: starting chain %04x\n",cardNum,chain->caid))
    chain->active=true;
    for(cPid *pid=chain->pids.First(); pid; pid=chain->pids.Next(pid)) {
      cPidFilter *filter=AddFilter(pid->pid,pid->sct,pid->mask,pid->mode,CHAIN_HOLD/8,false);
      if(filter) {
        filter->userData=(void *)chain;
        pid->filter=filter;
        }
      }
    }
}

void cLogger::StopChain(cLogChain *chain, bool force)
{
  if(chain->active) {
    if(force || (chain->delayed && chain->delay.TimedOut())) {
      dl(printf("logger %d: stopping chain %04x\n",cardNum,chain->caid))
      chain->active=false;
      for(cPid *pid=chain->pids.First(); pid; pid=chain->pids.Next(pid)) {
        cPidFilter *filter=pid->filter;
        if(filter) {
          DelFilter(filter);
          pid->filter=0;
          }
        }
      }
    else if(!chain->delayed) {
      dle(printf("logger %d: delaying chain %04x\n",cardNum,chain->caid))
      chain->delayed=true;
      chain->delay.Set(CHAIN_HOLD);
      }
    }
}

cPidFilter *cLogger::AddFilter(int Pid, int Section, int Mask, int Mode, int IdleTime, bool Crc)
{
  cPidFilter *filter=NewFilter(IdleTime);
  if(filter) {
    if(Pid>1) filter->SetBuffSize(32768);
    filter->Start(Pid,Section,Mask,Mode,Crc);
    dle(printf("logger %d: added filter pid=0x%.4x sct=0x%.2x/0x%.2x/0x%.2x idle=%d crc=%d\n",cardNum,Pid,Section,Mask,Mode,IdleTime,Crc))
    }
  else d(printf("logger %d: no free slot or filter failed to open\n",cardNum))
  return filter;
}

void cLogger::Process(cPidFilter *filter, unsigned char *data, int len)
{
  if(data && len>0) {
    if(filter==catfilt) {
      int vers=(data[5]&0x3E)>>1;
      if(data[0]==0x01 && vers!=catVers) {
        dle(printf("logger %d: got CAT version %02x\n",cardNum,vers))
        catVers=vers;
#ifdef LOG_CAT
        LogDump(data,len,"CAT vers %02x",catVers);
#endif
        ClearChains();
        for(int i=8; i<len-4; i+=data[i+1]+2) {
          if(data[i]==0x09) {
            int caid=WORD(data,i+2,0xFFFF);
            cLogChain *chain;
            for(chain=chains.First(); chain; chain=chains.Next(chain))
              if(chain->caid==caid) break;
            if(chain)
              chain->Parse(&data[i]);
            else {
              chain=new cLogChain(cardNum,softCSA);
              if(chain->Parse(&data[i]))
                chains.Add(chain);
              else
                delete chain;
              }
            }
          }
        SetChains();
        }
      }
    else {
#ifdef LOG_EMM
      LogDump(data,len,"EMM pid 0x%04x",filter->Pid());
#endif
#ifdef DEBUG_LOG_STATS
      if(logstats) logstats->Count();
#endif
      if(SCT_LEN(data)==len) {
        cLogChain *chain=(cLogChain *)(filter->userData);
        if(chain) {
          chain->Process(filter->Pid(),data);
          if(chain->delayed) StopChain(chain,false);
          }
        }
      else d(printf("logger %d: incomplete section %d != %d\n",cardNum,len,SCT_LEN(data)))
      }
    }
  else {
    cLogChain *chain=(cLogChain *)(filter->userData);
    if(chain && chain->delayed) StopChain(chain,false);
    }
}

// -- cEcmData -----------------------------------------------------------------

class cEcmData : public cEcmInfo {
private:
  bool del;
public:
  cEcmData(void);
  cEcmData(cEcmInfo *e);
  bool Save(FILE *f);
  bool Parse(const char *buf);
  void Delete(void) { del=true; }
  bool IsDeleted(void) const { return del; }
  };

cEcmData::cEcmData(void)
:cEcmInfo()
{
  del=false;
}

cEcmData::cEcmData(cEcmInfo *e)
:cEcmInfo(e)
{
  del=false;
}

bool cEcmData::Parse(const char *buf)
{
  char Name[64];
  int nu=0, num;
  Name[0]=0;
  if(sscanf(buf,"%d:%x:%x:%63[^:]:%x/%x:%x:%x/%x:%d%n",&prgId,&source,&transponder,Name,&caId,&emmCaId,&provId,&ecm_pid,&ecm_table,&nu,&num)>=9) {
    SetName(Name);
    const char *line=buf+num;
    if(nu>0 && *line++==':') {
      unsigned char dat[nu];
      if(GetHex(line,dat,nu,true)==nu) AddData(dat,nu);
      }
    return true;
    }
  return false;
}

bool cEcmData::Save(FILE *f)
{
  fprintf(f,"%d:%x:%x:%s:%x/%x:%x:%x/%x",prgId,source,transponder,name,caId,emmCaId,provId,ecm_pid,ecm_table);
  if(data) {
    char str[dataLen*2+2];
    fprintf(f,":%d:%s\n",dataLen,HexStr(str,data,dataLen));
    }
  else
    fprintf(f,":0\n");
  return ferror(f)==0;
}

// -- cEcmCache ----------------------------------------------------------------

cEcmCache ecmcache;

cEcmCache::cEcmCache(void)
:cLoader("ECM")
{}

void cEcmCache::New(cEcmInfo *e)
{
  Lock();
  cEcmData *dat;
  if(!(dat=Exists(e))) {
    dat=new cEcmData(e);
    Add(dat);
    Modified();
    d(printf("ecmcache: added prgId=%d source=%x transponder=%x ecm=%x/%x\n",e->prgId,e->source,e->transponder,e->ecm_pid,e->ecm_table))
    }
  else {
    if(strcasecmp(e->name,dat->name)) {
      dat->SetName(e->name);
      Modified();
      }
    if(dat->Update(e))
      Modified();
    }
  e->SetCached();
  Unlock();
}

cEcmData *cEcmCache::Exists(cEcmInfo *e)
{
  for(cEcmData *dat=First(); dat; dat=Next(dat))
    if(!dat->IsDeleted() && dat->Compare(e)) return dat;
  return 0;
}

int cEcmCache::GetCached(cSimpleList<cEcmInfo> *list, int sid, int Source, int Transponder)
{
  int n=0;
  list->Clear();
  Lock();
  for(cEcmData *dat=First(); dat; dat=Next(dat)) {
    if(!dat->IsDeleted() && dat->prgId==sid && dat->source==Source && dat->transponder==Transponder) {
      cEcmInfo *e=new cEcmInfo(dat);
      if(e) {
        d(printf("ecmcache: from cache: system %s (%04x) id %04x with ecm %x/%x\n",e->name,e->caId,e->provId,e->ecm_pid,e->ecm_table))
        e->SetCached();
        list->Add(e);
        n++;
        }
      }
    }
  Unlock();
  return n;
}

void cEcmCache::Delete(cEcmInfo *e)
{
  Lock();
  cEcmData *dat=Exists(e);
  if(dat) {
    dat->Delete();
    Modified();
    d(printf("ecmcache: invalidated prgId=%d source=%x transponder=%x ecm=%x/%x\n",dat->prgId,dat->source,dat->transponder,dat->ecm_pid,dat->ecm_table))
    }
  Unlock();
}

void cEcmCache::Flush(void)
{
  Lock();
  Clear();
  Modified();
  d(printf("ecmcache: flushed\n"))
  Unlock();    
}

void cEcmCache::Load(void)
{
  Lock();
  Clear();
  Unlock();
}

bool cEcmCache::Save(FILE *f)
{
  bool res=true;
  Lock();
  for(cEcmData *dat=First(); dat;) {
    if(!dat->IsDeleted()) {
      if(!dat->Save(f)) { res=false; break; }
      dat=Next(dat);
      }
    else {
      cEcmData *n=Next(dat);
      Del(dat);
      dat=n;
      }
    }
  Modified(!res);
  Unlock();
  return res;
}

bool cEcmCache::ParseLine(const char *line, bool fromCache)
{
  bool res=false;
  cEcmData *dat=new cEcmData;
  if(dat && dat->Parse(line)) {
    if(!Exists(dat)) { Add(dat); dat=0; }
    res=true;
    }
  delete dat;
  return res;
}

// -- cPrgPid ---------------------------------------------------------------------

class cPrgPid : public cSimpleItem {
private:
  int type;
  int pid;
  bool proc;
public:
  cPrgPid(int Type, int Pid);
  int Pid(void) { return pid; }
  int Type(void) { return type; }
  bool Proc(void) { return proc; }
  void Proc(bool is) { proc=is; };
  };

cPrgPid::cPrgPid(int Type, int Pid)
{
  type=Type; pid=Pid;
  proc=false;
}

// -- cPrg ---------------------------------------------------------------------

class cPrg : public cSimpleItem {
private:
  int prg;
  unsigned char cadescr[2048];
  int cadescrlen;
  bool streamflag, isUpdate;
public:
  cSimpleList<cPrgPid> pids;
  //
  cPrg(int Prg, bool IsUpdate);
  void AddCaDescr(const unsigned char *data, int len, bool flag);
  int GetCaDescr(unsigned char *buff, int max, bool &flag);
  int Prg(void) { return prg; }
  bool IsUpdate(void) { return isUpdate; }
  };

cPrg::cPrg(int Prg, bool IsUpdate)
{
  prg=Prg; isUpdate=IsUpdate;
  streamflag=false;
  cadescrlen=0;
}

void cPrg::AddCaDescr(const unsigned char *data, int len, bool flag)
{
  memcpy(cadescr,data,len);
  cadescrlen=len;
  streamflag=flag;
}

int cPrg::GetCaDescr(unsigned char *buff, int max, bool &flag)
{
  if(cadescrlen>max) return -1;
  if(cadescrlen==0) return 0;
  memcpy(buff,cadescr,cadescrlen);
  flag=streamflag;
  return cadescrlen;
}

// -- cEcmSys ------------------------------------------------------------------

class cEcmPri : public cSimpleItem {
public:
  cEcmInfo *ecm;
  int pri, sysIdent;
  };

// -- cEcmHandler --------------------------------------------------------------

class cEcmHandler : public cSimpleItem, public cAction {
private:
  int cardNum, cwIndex;
  cCam *cam;
  char *id;
  cTimeMs idleTime;
  //
  cMutex dataMutex;
  int sid;
  cSimpleList<cPrgPid> pids;
  unsigned char cadescr[512];
  int cadescrlen;
  //
  cSystem *sys;
  cPidFilter *filter;
  int filterCwIndex, filterSource, filterTransponder, filterSid;
  unsigned char lastCw[16];
  bool sync, noKey, trigger;
  int triggerMode;
  int mode, count;
  cTimeMs lastsync, startecm, resendTime;
  unsigned int cryptPeriod;
  unsigned char parity;
  cMsgCache failed;
  //
  cSimpleList<cEcmInfo> ecmList;
  cSimpleList<cEcmPri> ecmPriList;
  cEcmInfo *ecm;
  cEcmPri *ecmPri;
  cTimeMs ecmUpdTime;
  //
  int dolog;
  //
  void DeleteSys(void);
  void NoSync(bool clearParity);
  cEcmInfo *NewEcm(void);
  cEcmInfo *JumpEcm(void);
  void StopEcm(void);
  bool UpdateEcm(void);
  void EcmOk(void);
  void EcmFail(void);
  void ParseCAInfo(int sys);
  void AddEcmPri(cEcmInfo *n);
protected:
  virtual void Process(cPidFilter *filter, unsigned char *data, int len);
public:
  cEcmHandler(cCam *Cam, int CardNum, int cwindex);
  virtual ~cEcmHandler();
  void Stop(void);
  void SetPrg(cPrg *prg);
  void ShiftCwIndex(int cwindex);
  char *CurrentKeyStr(void) const;
  bool IsRemoveable(void);
  bool IsIdle(void);
  int Sid(void) const { return sid; }
  int CwIndex(void) const { return cwIndex; }
  const char *Id(void) const { return id; }
  };

cEcmHandler::cEcmHandler(cCam *Cam, int CardNum, int cwindex)
:cAction("ecmhandler",CardNum)
,failed(32,0)
{
  cam=Cam;
  cardNum=CardNum;
  cwIndex=cwindex;
  sys=0; filter=0; ecm=0; ecmPri=0; mode=-1; sid=-1; cadescrlen=0;
  trigger=false; triggerMode=-1;
  filterSource=filterTransponder=0; filterCwIndex=-1; filterSid=-1;
  asprintf(&id,"ecmhandler %d.%d",cardNum,cwindex);
}

cEcmHandler::~cEcmHandler()
{
  Lock();
  StopEcm();
  DelAllFilter(); // delete filters before sys for multi-threading reasons
  DeleteSys();
  Unlock();
  free(id);
}

bool cEcmHandler::IsIdle(void)
{
  dataMutex.Lock();
  int n=pids.Count();
  dataMutex.Unlock();
  return n==0;
}

bool cEcmHandler::IsRemoveable(void)
{
  return IsIdle() && idleTime.Elapsed()>MAX_ECM_IDLE;
}

void cEcmHandler::Stop(void)
{
  dataMutex.Lock();
  d(printf("%s: stop\n",id))
  sid=-1;
  idleTime.Set();
  pids.Clear();
  cadescrlen=0;
  trigger=true;
  dataMutex.Unlock();
  if(filter) filter->Wakeup();
}

void cEcmHandler::ShiftCwIndex(int cwindex)
{
  if(cwIndex!=cwindex) {
    dp(printf("%s: shifting cwIndex from %d to %d\n",id,cwIndex,cwindex))
    free(id);
    asprintf(&id,"ecmhandler %d.%d",cardNum,cwindex);
    dataMutex.Lock();
    trigger=true;
    cwIndex=cwindex;
    for(cPrgPid *pid=pids.First(); pid; pid=pids.Next(pid))
      cam->SetCWIndex(pid->Pid(),cwIndex);
    dataMutex.Unlock();
    if(filter) filter->Wakeup();
    }
}

void cEcmHandler::SetPrg(cPrg *prg)
{
  dataMutex.Lock();
  bool wasIdle=IsIdle();
  if(prg->Prg()!=sid) {
    d(printf("%s: setting new SID %d\n",id,prg->Prg()))
    sid=prg->Prg();
    idleTime.Set();
    pids.Clear();
    cadescrlen=0;
    trigger=true;
    }

#ifdef DEBUG_PIDS
  dp(printf("%s: pids on entry",id))
  for(cPrgPid *pid=pids.First(); pid; pid=pids.Next(pid))
    dp(printf(" %s=%04x",TYPENAME(pid->Type()),pid->Pid()))
  dp(printf("\n"))
#endif

  for(cPrgPid *pid=pids.First(); pid;) {
    cPrgPid *npid;
    for(npid=prg->pids.First(); npid; npid=prg->pids.Next(npid)) {
      if(pid->Pid()==npid->Pid()) {
        npid->Proc(true);
        break;
        }
      }
    if(!npid) {
      npid=pids.Next(pid);
      pids.Del(pid);
      pid=npid;
      }
    else pid=pids.Next(pid);
    }
#ifdef DEBUG_PIDS
  dp(printf("%s: pids after delete",id))
  for(cPrgPid *pid=pids.First(); pid; pid=pids.Next(pid))
    dp(printf(" %s=%04x",TYPENAME(pid->Type()),pid->Pid()))
  dp(printf("\n"))
#endif
  for(cPrgPid *npid=prg->pids.First(); npid; npid=prg->pids.Next(npid)) {
    if(!npid->Proc()) {
      cPrgPid *pid=new cPrgPid(npid->Type(),npid->Pid());
      pids.Add(pid);
      cam->SetCWIndex(pid->Pid(),cwIndex);
      }
    }
#ifdef DEBUG_PIDS
  dp(printf("%s: pids after add",id))
  for(cPrgPid *pid=pids.First(); pid; pid=pids.Next(pid))
    dp(printf(" %s=%04x",TYPENAME(pid->Type()),pid->Pid()))
  dp(printf("\n"))
#endif
  if(!IsIdle()) {
    trigger=true;
    triggerMode=0;
    if(wasIdle) d(printf("%s: is no longer idle\n",id))
    }
  else {
    if(!wasIdle) idleTime.Set();
    d(printf("%s: is idle%s\n",id,wasIdle?"":" now"))
    }

  int len;
  bool flag;
  len=prg->GetCaDescr(cadescr,sizeof(cadescr),flag);
  if(len>0) cadescrlen=len;

  if(!filter) {
    filter=NewFilter(IDLE_SLEEP);
    if(!filter) d(printf("%s: failed to open ECM filter\n",id))
    }
  dataMutex.Unlock();
  if(filter) filter->Wakeup();
}

void cEcmHandler::Process(cPidFilter *filter, unsigned char *data, int len)
{
  dataMutex.Lock();
  if(trigger) {
    d(printf("%s: triggered SID %d/%d idx %d/%d mode %d/%d %s\n",
      id,filterSid,sid,filterCwIndex,cwIndex,mode,triggerMode,(mode==3 && sync)?"sync":"-"))
    trigger=false;
    if(filterSid!=sid) {
      filterSid=sid;
      filterSource=cam->Source();
      filterTransponder=cam->Transponder();
      filterCwIndex=cwIndex;
      noKey=true; mode=0;
      }
    else {
      if(filterCwIndex!=cwIndex) {
        filterCwIndex=cwIndex;
        if(mode==3 && sync)
          cam->WriteCW(filterCwIndex,lastCw,true);
        }
      if(mode<triggerMode) mode=triggerMode;
      }
    triggerMode=-1;
    }
  dataMutex.Unlock();

  switch(mode) {
    case -1:
      filter->SetIdleTime(IDLE_SLEEP);
      break;

    case 0:
      StopEcm();
      if(IsIdle()) { mode=-1; break; }

      dolog=LOG_COUNT;
      NewEcm();
      filter->SetIdleTime(IDLE_GETCA);
      startecm.Set();
      mode=1;
      // fall through

    case 1:
      if(!ecm && !JumpEcm()) {
        if(startecm.Elapsed()>IDLE_GETCA_SLOW) {
          if(IsIdle()) { mode=0; break; }
          d(printf("%s: no encryption system found\n",id))
          filter->SetIdleTime(IDLE_GETCA_SLOW/4);
          startecm.Set();
          }
        break;
        }
      mode=4;
      // fall through

    case 4:
    case 5:
      NoSync(mode==4);
      failed.Clear();
      filter->SetIdleTime(IDLE_NO_SYNC/2);
      lastsync.Set();
      cryptPeriod=60*1000;
      mode=2;
      // fall through
          
    case 2:
      if(sys->NeedsData()) {
        if(!UpdateEcm()) {
          if(lastsync.Elapsed()<ECM_DATA_TIME) break;
          d(printf("%s: no ecm extra data update (waited %lld ms)\n",id,lastsync.Elapsed()))
          }
        if(lastsync.Elapsed()>IDLE_NO_SYNC/4 && dolog)
          d(printf("%s: ecm extra data update took %lld ms\n",id,lastsync.Elapsed()))
        }
      filter->SetIdleTime(IDLE_NO_SYNC);
      mode=3;
      // fall through

    case 3:
      {
      bool resend=false, cwok=false;
      if(resendTime.TimedOut()) {
        resend=sync; resendTime.Set(8*24*60*60*1000);
        }

#ifdef DEBUG_AV7110
      if(startecm.Elapsed()<3*60*1000) cam->DumpAV7110();
#endif

      if(data && len>0) {
#ifdef LOG_ECM
        LogDump(data,len,"ECM sys 0x%04x id 0x%02x pid 0x%04x",ecm->caId,ecm->provId,filter->Pid());
#endif

        if(SCT_LEN(data)==len) {
          depDUMP(data,16,"%s: ECM",id);
          int n;
          if(!(n=sys->CheckECM(ecm,data,sync))) {
            if(resend || parity!=(data[0]&1)) {
              int ecmid;
              cTimeMs procTime;
              cwok=(ecmid=failed.Get(data,len,0))>=0 && sys->ProcessECM(ecm,data);
              n=(ecmid>0)?failed.Cache(ecmid,cwok,0):99;
              sys->CheckECMResult(ecm,data,cwok);
              if(cwok) {
                parity=data[0]&1;
                }
              else {
                if(procTime.Elapsed()>6000) {
                  d(printf("%s: filter flush (elapsed %d)\n",id,(int)procTime.Elapsed()))
                  filter->Flush();
                  }
                if(n>=2) { count++; if(n==2) count++; }
                parity=0xFF;
                if(sync && lastsync.Elapsed()>cryptPeriod*2) {
                  d(printf("%s: lost sync (period %d, elapsed %d)\n",id,cryptPeriod,(int)lastsync.Elapsed()))
                  NoSync(true);
                  }
                }
              dep(printf("%s: ECM proc (%s) cwok=%d ecmid=%d n=%d sync=%d parity=%d count=%d ELA=%d\n",
                    id,sys->Name(),cwok,ecmid,n,sync,parity,count,(int)procTime.Elapsed()))
              }
            }
          else {
            dep(printf("%s: ECM check result %d\n",id,n))
            switch(n) {
              case 1: NoSync(true); break;
              case 2: count++; break;
              case 3: break;
              }
            }
          }
        else {
          d(printf("%s: incomplete section %d != %d\n",id,len,SCT_LEN(data)))
          count++;
          }
        }
      else if(sys->Constant()) {
        if(sys->ProcessECM(ecm,NULL)) {
          cwok=true;
          if(sync) filter->SetIdleTime(IDLE_SYNC*10); 
          }
        else count++;
        }
      else count++;

      if(cwok) {
        dolog=LOG_COUNT; sys->DoLog(true);
        cam->WriteCW(filterCwIndex,sys->CW(),resend || !sync);
        memcpy(lastCw,sys->CW(),sizeof(lastCw));
        noKey=false; count=0;
        UpdateEcm(); EcmOk();
        if(!sync) {
          sync=true;
          filter->SetIdleTime(IDLE_SYNC);
          d(printf("%s: correct key found\n",id))
          if(!cam->IsSoftCSA())
            resendTime.Set(CW_REPEAT_TIME);
          }
        else if(!resend)
          cryptPeriod=max(5000,min(60000,(int)lastsync.Elapsed()));
        lastsync.Set();
        }

      if(!sync && !trigger) {
        if(count>=sys->MaxEcmTry()) {
          EcmFail(); JumpEcm();
          mode=4;
          if(!ecm) {
            JumpEcm();
            if(!ecm) { // this should not happen!
              d(printf("%s: BUG: empty ecm list in sync loop\n",id))
              mode=0; break;
              }
            // if we looped through all systems, we wait until the next parity
            // change before we try again.
            if(dolog!=LOG_COUNT && data) { parity=data[0]&1; mode=5; }
            if(dolog && !--dolog) {
#ifdef DEBUG
              sys->DoLog(false);
              printf("%s: stopping console log until valid key is found\n",id);
#endif
              }
            }
          break;
          }
        }
        
      if(IsIdle() && idleTime.Elapsed()>MAX_ECM_HOLD) {
        d(printf("%s: hold timeout expired\n",id))
        mode=0;
        }

      break;
      }
    }
}

void cEcmHandler::NoSync(bool clearParity)
{
  if(clearParity) parity=0xFF;
  count=0; sync=false;
}

void cEcmHandler::DeleteSys(void)
{
  delete sys; sys=0;
}

char *cEcmHandler::CurrentKeyStr(void) const
{
  if(noKey || !sys) return 0;
  return strdup(sys->CurrentKeyStr());
}

cEcmInfo *cEcmHandler::NewEcm(void)
{
  ecmcache.GetCached(&ecmList,filterSid,filterSource,filterTransponder);
  ecmPriList.Clear();
  for(cEcmInfo *n=ecmList.First(); n; n=ecmList.Next(n)) AddEcmPri(n);
  ecm=0; ecmPri=0;
  return JumpEcm();
}

void cEcmHandler::AddEcmPri(cEcmInfo *n)
{
  int ident, pri=0;
  while((ident=cSystems::FindIdentBySysId(n->caId,!cam->IsSoftCSA(),pri))>0) {
    cEcmPri *ep=new cEcmPri;
    if(ep) {
      ep->ecm=n; 
      ep->pri=pri;
      ep->sysIdent=ident;
      if(n->Cached() && (!ScSetup.LocalPriority || pri!=-15)) ep->pri+=20;

      // keep ecmPriList sorted
      cEcmPri *eppp, *epp=ecmPriList.First();
      if(!epp || epp->pri<ep->pri)
        ecmPriList.Ins(ep);
      else {
        do {
          eppp=ecmPriList.Next(epp);
          if(!eppp || eppp->pri<ep->pri) {
            ecmPriList.Add(ep,epp);
            break;
            }
          } while((epp=eppp));
        }
      }
    }
}

void cEcmHandler::StopEcm(void)
{
  filter->Stop(); filter->Flush();
  if(ecm) cam->LogEcmStatus(ecm,false);
  DeleteSys();
}

bool cEcmHandler::UpdateEcm(void)
{
  if(!ecm->Data() || ecmUpdTime.TimedOut()) {
#ifdef DEBUG
    bool log=dolog;
    dolog=(sys && sys->NeedsData() && ecm->Data()==0);
    if(dolog) printf("%s: try to update ecm extra data\n",id);
#endif
    ParseCAInfo(ecm->caId);
    ecmUpdTime.Set(ECM_UPD_TIME);
#ifdef DEBUG
    dolog=log;
#endif
    if(!ecm->Data()) return false;
    }
  return true;
}

cEcmInfo *cEcmHandler::JumpEcm(void)
{
  noKey=true;
  if(!ecmPri) {
    ParseCAInfo(0xFFFF); // all systems
    ecmPri=ecmPriList.First();
    }
  else ecmPri=ecmPriList.Next(ecmPri);
  if(ecmPri) {
    if(ecmPri->ecm!=ecm) {
      StopEcm(); ecmUpdTime.Set();
      ecm=ecmPri->ecm;
      filter->Start(ecm->ecm_pid,ecm->ecm_table,0xfe,0,false);
      cam->LogEcmStatus(ecm,true);
      }
    else {
      DeleteSys();
      filter->Flush();
      }
    sys=cSystems::FindBySysIdent(ecmPri->sysIdent);
    if(!sys) {
      if(dolog) d(printf("%s: internal error: no system found for ident %04x (caid %04x pri %d)\n",id,ecmPri->sysIdent,ecmPri->ecm->caId,ecmPri->pri))
      return JumpEcm();
      }
    sys->DoLog(dolog!=0); sys->CardNum(cardNum);
    failed.SetMaxFail(sys->MaxEcmTry());

    if(dolog) d(printf("%s: try system %s (%04x) id %04x with ecm %x%s (pri=%d)\n",
                     id,sys->Name(),ecm->caId,ecm->provId,ecm->ecm_pid,ecm->Cached()?" (cached)":"",sys->Pri()))
    }
  else {
    StopEcm();
    ecm=0;
    }
  return ecm;
}

void cEcmHandler::EcmOk(void)
{
  ecm->SetName(sys->Name());
  ecm->Fail(false);
  ecmcache.New(ecm);
  cEcmInfo *e=ecmList.First();
  while(e) {
    if(e->Cached() && e->Failed()) ecmcache.Delete(e);
    e=ecmList.Next(e);
    }
  cSoftCAM::Save();
}

void cEcmHandler::EcmFail(void)
{
  ecm->Fail(true);
}

void cEcmHandler::ParseCAInfo(int SysId)
{
  if(cadescrlen>0) {
    if(dolog) d(printf("%s: got CaDescriptors for SID %04X (len=%d)\n",id,sid,cadescrlen))
#ifdef LOG_PMT
    LogDump(cadescr,cadescrlen,"PMT");
#endif
    for(int index=0; index<cadescrlen; index+=cadescr[index+1]+2) {
      if(cadescr[index]==0x09) {
#ifdef DEBUG
        if(dolog) {
          printf("%s: descriptor",id);
          for(int i=0 ; i<cadescr[index+1] ; i++) printf(" %02x",cadescr[index+2+i]);
          printf("\n");
          }
#endif
        int sysId=WORD(cadescr,index+2,0xFFFF);
        if(SysId!=0xFFFF && sysId!=SysId) continue;
        int sysPri=0;
        cSystem *sys;
        while((sys=cSystems::FindBySysId(sysId,!cam->IsSoftCSA(),sysPri))) {
          sysPri=sys->Pri();
          cSimpleList<cEcmInfo> ecms;
          sys->ParseCADescriptor(&ecms,sysId,&cadescr[index+2],cadescr[index+1]);
          delete sys;
          if(ecms.Count()) {
            cEcmInfo *n;
            while((n=ecms.First())) {
              ecms.Del(n,false);
              if(dolog) d(printf("%s: found %04x (%s) id %04x with ecm %x ",id,n->caId,n->name,n->provId,n->ecm_pid))
              cEcmInfo *e=ecmList.First();
              while(e) {
                if(e->ecm_pid==n->ecm_pid) {
                  if(e->caId==n->caId && e->provId==n->provId) {
                    if(n->Data()) {
                      if(e->Update(n) && dolog) d(printf("(updated) "))
                      }
                    if(dolog) d(printf("(already present)\n"))
                    delete n; n=0;
                    break;
                    }
                  else {
                    e->Fail(true);
                    if(dolog) d(printf("(dup) "))
                    }
                  }
                e=ecmList.Next(e);
                }
              if(n) {
                if(dolog) d(printf("(new)\n"))
                n->SetSource(sid,filterSource,filterTransponder);
                ecmList.Add(n);
                AddEcmPri(n);
                }
              }
            break;
            }
          }
        if(sysPri==0 && dolog) d(printf("%s: no module available for system %04x\n",id,sysId))
        }
      }
    }
#ifdef DEBUG_ECM_PROG
  if(SysId==0xffff) {
    for(cEcmPri *ep=ecmPriList.First(); ep; ep=ecmPriList.Next(ep))
      printf("%s: ecmPriList pri=%d ident=%04x caid=%04x pid=%04x\n",id,ep->pri,ep->sysIdent,ep->ecm->caId,ep->ecm->ecm_pid);
    printf("%s: ecmPri list end\n",id);
    }
#endif
}

// -- cCam ---------------------------------------------------------------

cCam::cCam(cScDvbDevice *dev, int CardNum)
{
  device=dev; cardNum=CardNum;
  source=transponder=-1; liveVpid=liveApid=0; logger=0;
  memset(lastCW,0,sizeof(lastCW));
  memset(indexMap,0,sizeof(indexMap));
}

cCam::~cCam()
{
  handlerList.Clear();
  delete logger;
}

bool cCam::IsSoftCSA(void)
{
  return device->SoftCSA();
}

void cCam::Tune(const cChannel *channel)
{
  cMutexLock lock(this);
  if(source!=channel->Source() || transponder!=channel->Transponder()) {
    source=channel->Source(); transponder=channel->Transponder();
    dp(printf("cam %d: now tuned to source %x transponder %x\n",cardNum,source,transponder))
    Stop();
    }
  else dp(printf("cam %d: tune to same source/transponder\n",cardNum))
}

void cCam::SetPid(int type, int pid, bool on)
{
  cMutexLock lock(this);
#ifdef DEBUG_PIDS
  int oldA=liveApid, oldV=liveVpid;
#endif
  if(type==1) liveVpid=on ? pid:0;
  else if(type==0) liveApid=on ? pid:0;
  else if(liveVpid==pid && on) liveVpid=0;
  else if(liveApid==pid && on) liveApid=0;
#ifdef DEBUG_PIDS
  if(oldA!=liveApid || oldV!=liveVpid)
    printf("cam %d: livepids video=%04x audio=%04x\n",cardNum,liveVpid,liveApid);
#endif
}

void cCam::Stop(void)
{
  for(cEcmHandler *handler=handlerList.First(); handler; handler=handlerList.Next(handler))
    handler->Stop();
  if(logger) logger->Down();
}

void cCam::AddPrg(cPrg *prg)
{
  cMutexLock lock(this);
  bool islive=false;
  for(cPrgPid *pid=prg->pids.First(); pid; pid=prg->pids.Next(pid))
    if(pid->Pid()==liveVpid || pid->Pid()==liveApid) {
      islive=true;
      break;
      }
  bool needZero=!IsSoftCSA() && (islive || !ScSetup.ConcurrentFF);
  bool noshift=IsSoftCSA() || prg->IsUpdate();
  dp(printf("cam %d: %s prg %04x (zero=%d noshift=%d)\n",cardNum,prg->IsUpdate()?"update":"add",prg->Prg(),needZero,noshift))
#ifdef DEBUG_PIDS
  if(prg->pids.Count()>0) {
    dp(printf("cam %d: pids",cardNum))
    for(cPrgPid *pid=prg->pids.First(); pid; pid=prg->pids.Next(pid))
      dp(printf(" %s=%04x",TYPENAME(pid->Type()),pid->Pid()))
    dp(printf("\n"))
    }
#endif
  cEcmHandler *handler=GetHandler(prg->Prg(),needZero,noshift);
  if(handler) {
    dp(printf("cam %d: found handler for SID %d (%s idle=%d idx=%d)\n",cardNum,prg->Prg(),handler->Id(),handler->IsIdle(),handler->CwIndex()))
    handler->SetPrg(prg);
    }
}

char *cCam::CurrentKeyStr(int num)
{
  cMutexLock lock(this);
  cEcmHandler *handler;
  for(handler=handlerList.First(); handler; handler=handlerList.Next(handler))
    if(--num<0) return handler->CurrentKeyStr();
  return 0;
}

bool cCam::Active(void)
{
  cMutexLock lock(this);
  for(cEcmHandler *handler=handlerList.First(); handler; handler=handlerList.Next(handler))
    if(!handler->IsIdle()) {
      d(printf("cam %d: %s non-idle\n",cardNum,handler->Id()))
      return true;
      }
  return false;
}

void cCam::HouseKeeping(void)
{
  cMutexLock lock(this);
  for(cEcmHandler *handler=handlerList.First(); handler;) {
    cEcmHandler *next=handlerList.Next(handler);
    if(handler->IsRemoveable()) RemHandler(handler);
    handler=next;
    }
  if(handlerList.Count()<1) {
    delete logger; logger=0;
    }
}

void cCam::LogEcmStatus(const cEcmInfo *ecm, bool on)
{
  cMutexLock lock(this);
  if(!logger && on && ScSetup.AutoUpdate) {
    logger=new cLogger(cardNum,IsSoftCSA());
#ifdef DEBUG_LOG_STATS
    LogStatsUp();
#endif
    }
  if(logger) logger->EcmStatus(ecm,on);
}

void cCam::SetCWIndex(int pid, int index)
{
  if(index<MAX_CW_IDX) {
    ca_pid_t ca_pid;
    ca_pid.pid=pid;
    ca_pid.index=index;
    dp(printf("cam %d: descrambling pid %04x on index %x\n",cardNum,pid,index))
    if(!device->SetCaPid(&ca_pid))
      if(index>0) {
        LogError("CA_SET_PID failed (%s). Expect a black screen/bad recording. Do you use the patched DVB driver?",strerror(errno));
        LogWarn("Adjusting 'Concurrent FF streams' to NO");
        ScSetup.ConcurrentFF=0;
        cSoftCAM::StoreConfig(true);
        }
    }
}

void cCam::WriteCW(int index, unsigned char *cw, bool force)
{
  if(index<MAX_CW_IDX) {
    for(int i=0; i<16; i+=4) cw[i+3]=cw[i]+cw[i+1]+cw[i+2];
    ca_descr_t ca_descr;
    ca_descr.index=index;
    unsigned char *last=lastCW[index];
    if(force || memcmp(&cw[0],&last[0],8)) {
      memcpy(&last[0],&cw[0],8);
      ca_descr.parity=0;
      memcpy(ca_descr.cw,&cw[0],8);
      if(!device->SetCaDescr(&ca_descr))
        LogError("CA_SET_DESCR failed (%s). Expect a black screen.",strerror(errno));
      }

    if(force || memcmp(&cw[8],&last[8],8)) {
      memcpy(&last[8],&cw[8],8);
      ca_descr.parity=1;
      memcpy(ca_descr.cw,&cw[8],8);
      if(!device->SetCaDescr(&ca_descr))
        LogError("CA_SET_DESCR failed (%s). Expect a black screen.",strerror(errno));
      }
    }
}

void cCam::DumpAV7110(void)
{
  device->DumpAV7110();
}

int cCam::GetFreeIndex(void)
{
  for(int idx=0; idx<MAX_CW_IDX; idx++)
    if(!indexMap[idx]) return idx;
  return -1;
}

cEcmHandler *cCam::GetHandler(int sid, bool needZero, bool noshift)
{
  cEcmHandler *zerohandler=0, *sidhandler=0, *idlehandler=0;
  for(cEcmHandler *handler=handlerList.First(); handler; handler=handlerList.Next(handler)) {
    if(handler->Sid()==sid)
      sidhandler=handler;
    if(handler->CwIndex()==0)
      zerohandler=handler;
    if(handler->IsIdle() && (!idlehandler || (!(needZero ^ (idlehandler->CwIndex()!=0)) && (needZero ^ (handler->CwIndex()!=0))) ))
      idlehandler=handler;
    }
#ifdef DEBUG_PIDS
  printf("cam %d: SID=%d zero=%d |",cardNum,sid,needZero);
  if(sidhandler) printf(" sid=%d/%d/%d",sidhandler->CwIndex(),sidhandler->Sid(),sidhandler->IsIdle());
  else printf(" sid=-/-/-");
  if(zerohandler) printf(" zero=%d/%d/%d",zerohandler->CwIndex(),zerohandler->Sid(),zerohandler->IsIdle());
  else printf(" zero=-/-/-");
  if(idlehandler) printf(" idle=%d/%d/%d",idlehandler->CwIndex(),idlehandler->Sid(),idlehandler->IsIdle());
  else printf(" idle=-/-/-");
  printf("\n");
#endif

  if(sidhandler) {
    if(needZero && sidhandler->CwIndex()!=0 && !noshift) {
      if(!sidhandler->IsIdle())
        d(printf("cam %d: shifting cwindex on non-idle handler.\n",cardNum))
      if(zerohandler) {
        if(!zerohandler->IsIdle())
          d(printf("cam %d: shifting non-idle zero handler. This shouldn't happen!\n",cardNum))
        zerohandler->ShiftCwIndex(sidhandler->CwIndex());
        sidhandler->ShiftCwIndex(0);
        }
      else if(indexMap[0]==0) {
        indexMap[0]=1;
        indexMap[sidhandler->CwIndex()]=0;
        sidhandler->ShiftCwIndex(0);
        }
      else d(printf("cam %d: zero index not free.\n",cardNum))
      }

    if(!needZero && sidhandler->CwIndex()==0 && !noshift) {
      if(!sidhandler->IsIdle())
        d(printf("cam %d: shifting cwindex on non-idle handler.\n",cardNum))
      int idx=GetFreeIndex();
      if(idx>=0) {
        indexMap[idx]=1;
        sidhandler->ShiftCwIndex(idx);
        indexMap[0]=0;
        }
      else d(printf("cam %d: no free cwindex. Can't free zero index.\n",cardNum))
      }

    return sidhandler;
    }

  if(needZero && zerohandler) {
    if(!zerohandler->IsIdle())
      d(printf("cam %d: changing SID on non-idle zero handler. This shouldn't happen!\n",cardNum))
    return zerohandler;
    }
  
  if(idlehandler) {
    if(needZero && idlehandler->CwIndex()!=0 && !noshift) {
      if(indexMap[0]==0) {
        indexMap[0]=1;
        indexMap[idlehandler->CwIndex()]=0;
        idlehandler->ShiftCwIndex(0);
        }
      else d(printf("cam %d: zero index not free. (2)\n",cardNum))
      }
    if(!needZero && idlehandler->CwIndex()==0 && !noshift) {
      int idx=GetFreeIndex();
      if(idx>=0) {
        indexMap[idx]=1;
        idlehandler->ShiftCwIndex(idx);
        indexMap[0]=0;
        }
      else d(printf("cam %d: no free cwindex. Can't free zero index. (2)\n",cardNum))
      }
    if((needZero ^ (idlehandler->CwIndex()==0)))
      d(printf("cam %d: idlehandler index doesn't match needZero\n",cardNum))
    return idlehandler;
    }

  int idx=GetFreeIndex();
  if(!needZero && idx==0) {
    indexMap[0]=1;
    idx=GetFreeIndex();
    indexMap[0]=0;
    if(idx<0) {
      idx=0;
      d(printf("cam %d: can't respect !needZero for new handler\n",cardNum))
      }
    }
  if(idx<0) {
    d(printf("cam %d: no free cwindex\n",cardNum))
    return 0;
    }
  indexMap[idx]=1;
  idlehandler=new cEcmHandler(this,cardNum,idx);
  handlerList.Add(idlehandler);
  return idlehandler;
}

void cCam::RemHandler(cEcmHandler *handler)
{
  int idx=handler->CwIndex();
  dp(printf("cam %d: removing %s on cw index %d\n",cardNum,handler->Id(),idx))
  handlerList.Del(handler);
  indexMap[idx]=0;
}

// -- cScCiAdapter -------------------------------------------------------------

struct TPDU {
  unsigned char slot;
  unsigned char tcid;
  unsigned char tag;
  unsigned char len;
  unsigned char data[1];
  };

class cScCamSlot;

class cScCiAdapter : public cCiAdapter, private cMutex {
private:
  cDevice *device;
  cCam *cam;
  int cardIndex;
  cRingBufferLinear *rb;
  cScCamSlot *slots[MAX_CI_SLOTS];
  //
  int tcid;
protected:
  virtual int Read(unsigned char *Buffer, int MaxLength);
  virtual void Write(const unsigned char *Buffer, int Length);
  virtual bool Reset(int Slot);
  virtual eModuleStatus ModuleStatus(int Slot);
  virtual bool Assign(cDevice *Device, bool Query=false);
public:
  cScCiAdapter(cDevice *Device, int CardIndex, cCam *Cam);
  ~cScCiAdapter();
  void CamStop(void);
  void CamAddPrg(cPrg *prg);
  bool CamSoftCSA(void);
  void Shutdown(void);
  };

// -- cScCamSlot ---------------------------------------------------------------

#define SLOT_CAID_CHECK 120000

class cScCamSlot : public cCamSlot, public cRingBufferLinear {
private:
  cScCiAdapter *ciadapter;
  unsigned short caids[MAX_CI_SLOT_CAIDS+1];
  int slot, version;
  cTimeMs checkTimer;
  //
  int GetLength(const unsigned char * &data);
  void CaInfo(unsigned char *b, int tcid, int cid);
  bool Check(void);
public:
  cScCamSlot(cScCiAdapter *ca, int Slot);
  void Process(const unsigned char *data, int len);
  eModuleStatus Status(void);
  };

cScCamSlot::cScCamSlot(cScCiAdapter *ca, int Slot)
:cCamSlot(ca)
,cRingBufferLinear(KILOBYTE(2),5+1)
{
  ciadapter=ca; slot=Slot;
  version=cSoftCAM::GetCaids(slot,caids,MAX_CI_SLOT_CAIDS);
  checkTimer.Set(SLOT_CAID_CHECK);
}

eModuleStatus cScCamSlot::Status(void)
{
  return caids[0] ? msReady:msNone;
}

bool cScCamSlot::Check(void)
{
  if(checkTimer.TimedOut() && version!=cSoftCAM::GetCaids(slot,0,0)) {
    version=cSoftCAM::GetCaids(slot,caids,MAX_CI_SLOT_CAIDS);
    checkTimer.Set(SLOT_CAID_CHECK);
    return true;
    }
  return false;
}

int cScCamSlot::GetLength(const unsigned char * &data)
{
  int len=*data++;
  if(len&0x80)
    for(int i=len&~0x80, len=0; i>0; i--) len=(len<<8) + *data++;
  return len;
}

void cScCamSlot::CaInfo(unsigned char *b, int tcid, int cid)
{
  b[0]=0xa0; b[2]=tcid;
  b[3]=0x90;
  b[4]=0x02; b[5]=cid<<8; b[6]=cid&0xff;
  b[7]=0x9f; b[8]=0x80;   b[9]=0x31; // AOT_CA_INFO
  int l=0;
  for(int i=0; caids[i]; i++) {
    b[l+11]=caids[i]>>8;
    b[l+12]=caids[i]&0xff;
    l+=2;
    }
  b[10]=l; b[1]=l+9; b[-1]=l+11; Put(b-1,l+12);
}

void cScCamSlot::Process(const unsigned char *data, int len)
{
  data+=3;
  int dlen=GetLength(data);
  int tcid=data[0];

  unsigned char a[128], *b=&a[1];
  if(Check()) CaInfo(b,tcid,0x01);

  if(dlen<8 || data[1]!=0x90) return;
  int cid=(data[3]<<8)+data[4];
  int tag=(data[5]<<16)+(data[6]<<8)+data[7];
  data+=8;
  dlen=GetLength(data);
  switch(tag) {
    case 0x9f8030: // AOT_CA_INFO_ENQ
      CaInfo(b,tcid,cid);
      break;
    
    case 0x9f8032: // AOT_CA_PMT
      if(dlen>=6) {
        int ca_lm=data[0];
        int ci_cmd=-1;
        cPrg *prg=new cPrg((data[1]<<8)+data[2],ca_lm==5);
        int ilen=(data[4]<<8)+data[5];
        data+=6; dlen-=6;
        if(ilen>0 && dlen>=ilen) {
          ci_cmd=data[0];
          prg->AddCaDescr(&data[1],ilen-1,false);
          }
        data+=ilen; dlen-=ilen;
        while(dlen>=5) {
          cPrgPid *pid=new cPrgPid(data[0],(data[1]<<8)+data[2]);
          prg->pids.Add(pid);
          ilen=(data[3]<<8)+data[4];
          data+=5; dlen-=5;
          if(ilen>0 && dlen>=ilen) {
            ci_cmd=data[0];
            prg->AddCaDescr(&data[1],ilen-1,true);
            }
          data+=ilen; dlen-=ilen;
          }
        if((ciadapter->CamSoftCSA() || ScSetup.ConcurrentFF==1) && (ci_cmd==0x03 || (ci_cmd==0x01 && ca_lm==0x03))) {
          b[0]=0xa0; b[2]=tcid;
          b[3]=0x90;
          b[4]=0x02; b[5]=cid<<8; b[6]=cid&0xff;
          b[7]=0x9f; b[8]=0x80; b[9]=0x33; // AOT_CA_PMT_REPLY
          b[11]=prg->Prg()<<8;
          b[12]=prg->Prg()&0xff;
          b[13]=0x00;
          b[14]=0x81; 	// CA_ENABLE
          b[10]=4; b[1]=4+9; a[0]=4+11; Put(a,4+12);
          }
        if(ci_cmd==0x04) {
          ciadapter->CamStop();
          }
        if(ci_cmd==0x01) {
          ciadapter->CamAddPrg(prg);
          }
        delete prg;
        }
      break;
    }
}

// -- cScCiAdapter -------------------------------------------------------------

cScCiAdapter::cScCiAdapter(cDevice *Device, int CardIndex, cCam *Cam)
{
  device=Device; cardIndex=CardIndex; cam=Cam;
  tcid=0;
  memset(slots,0,sizeof(slots));
  SetDescription("SC-CI adapter on device %d",cardIndex);
  rb=new cRingBufferLinear(KILOBYTE(5),6+1);
  if(rb) {
    rb->SetTimeouts(0,CAM_READ_TIMEOUT);
    bool spare=true;
    for(int i=0; i<MAX_CI_SLOTS; i++) {
      if(cSoftCAM::GetCaids(i,0,0)<=0) {
        if(!spare) break;
        spare=false;
        }
      slots[i]=new cScCamSlot(this,i);
      }
    Start();
    }
  else di(printf("sc-ci %d: failed to create ringbuffer.\n",cardIndex))
}

cScCiAdapter::~cScCiAdapter()
{
  Shutdown();
  Cancel(3);
  delete rb;
}

void cScCiAdapter::Shutdown(void)
{
  cMutex::Lock(); cam=0; cMutex::Unlock();
}

void cScCiAdapter::CamStop(void)
{
  if(cam) cam->Stop();
}

void cScCiAdapter::CamAddPrg(cPrg *prg)
{
  if(cam) cam->AddPrg(prg);
}

bool cScCiAdapter::CamSoftCSA(void)
{
  return cam && cam->IsSoftCSA();
}

int cScCiAdapter::Read(unsigned char *Buffer, int MaxLength)
{
  cMutexLock lock(this);
  if(cam && rb && Buffer && MaxLength>0) {
    int c;
    unsigned char *data=rb->Get(c);
    if(data) {
      int s=data[0];
      if(c>=s) {
        c=s>MaxLength ? MaxLength : s;
        memcpy(Buffer,&data[1],c);
        if(c<s) { data[c]=s-c; rb->Del(c); }
        else rb->Del(c+1);
        return c;
        }
      else di(printf("sc-ci %d: internal error (rb frame sync)\n",cardIndex))
      }
    }
  return 0;
}

#define TPDU(data,slot)do { unsigned char *_d=(data); _d[0]=(slot); _d[1]=tcid; } while(0);
#define TAG(data,tag,len) do { unsigned char *_d=(data); _d[0]=(tag); _d[1]=(len); } while(0);
#define SB_TAG(data,sb) do { unsigned char *_d=(data); _d[0]=0x80; _d[1]=0x02; _d[2]=tcid; _d[3]=(sb); } while(0);
#define PUT_TAG(data,len) do { unsigned char *_d=(data)-1; int _l=(len); _d[0]=_l; if(rb) rb->Put(_d,_l+1); } while(0);

void cScCiAdapter::Write(const unsigned char *buff, int len)
{
  cMutexLock lock(this);
  if(cam && buff && len>=5) {
    unsigned char a[128], *b=&a[1];
    struct TPDU *tpdu=(struct TPDU *)buff;
    int slot=tpdu->slot;
    if(slots[slot]) {
      switch(tpdu->tag) {
        case 0x81: // T_RCV
          {
          TPDU(b,slot);
          int l=2, c;
          unsigned char *d=slots[slot]->Get(c);
          if(d) {
            int s=d[0];
            if(c>=s) {
              memcpy(&b[l],&d[1],s);
              l+=s;
              slots[slot]->Del(s+1);
              }
            else slots[slot]->Del(c);
            }
          SB_TAG(&b[l],0x00);
          PUT_TAG(b,l+4);
          break;
          }
        case 0x82: // T_CREATE_TC
          tcid=tpdu->data[0];
          TPDU(b,slot);
          TAG(&b[2],0x83,0x01); b[4]=tcid;
          SB_TAG(&b[5],0x00);
          PUT_TAG(b,9);

          static const unsigned char reqCAS[] = { 0xA0,0x07,0x01,0x91,0x04,0x00,0x03,0x00,0x41 };
          memcpy(b,reqCAS,sizeof(reqCAS));
          b[2]=tcid;
          a[0]=sizeof(reqCAS);
          slots[slot]->Put(a,sizeof(reqCAS)+1);
          break;
        case 0xA0: // T_DATA_LAST
          slots[slot]->Process(buff,len);
          TPDU(b,slot);
          SB_TAG(&b[2],slots[slot]->Available()>0 ? 0x80:0x00);
          PUT_TAG(b,6);
          break;
        }
      }
    }
}

bool cScCiAdapter::Reset(int Slot)
{
  return true;
}

eModuleStatus cScCiAdapter::ModuleStatus(int Slot)
{
  bool enable=cSoftCAM::CapCheck(cardIndex);
  if(!enable) CamStop();
  return (enable && cam) ? slots[Slot]->Status():msNone;
}

bool cScCiAdapter::Assign(cDevice *Device, bool Query)
{
  return Device ? (Device==device) : true;
}

// -- cDeCSA -------------------------------------------------------------------

#define MAX_CSA_PIDS 8192
#define MAX_CSA_IDX  16

class cDeCSA : public cMutex {
private:
  int cs;
  unsigned char **range;
  unsigned char pidmap[MAX_CSA_PIDS];
  void *keys[MAX_CSA_IDX];
  //
  bool GetKeyStruct(int idx);
public:
  cDeCSA(void);
  ~cDeCSA();
  bool Decrypt(unsigned char *data, int len, bool force);
  bool SetDescr(ca_descr_t *ca_descr);
  bool SetCaPid(ca_pid_t *ca_pid);
  };

cDeCSA::cDeCSA(void)
{
  cs=get_suggested_cluster_size();
  dc(printf("csa: clustersize=%d rangesize=%d\n",cs,cs*2+5))
  range=MALLOC(unsigned char *,(cs*2+5));
  memset(keys,0,sizeof(keys));
  memset(pidmap,0,sizeof(pidmap));
}

cDeCSA::~cDeCSA()
{
  for(int i=0; i<MAX_CSA_IDX; i++)
    if(keys[i]) free_key_struct(keys[i]);
  free(range);
}

bool cDeCSA::GetKeyStruct(int idx)
{
  if(!keys[idx]) keys[idx]=get_key_struct();
  return keys[idx]!=0;
}

bool cDeCSA::SetDescr(ca_descr_t *ca_descr)
{
  cMutexLock lock(this);
  int idx=ca_descr->index;
  if(idx<MAX_CSA_IDX && GetKeyStruct(idx)) {
    if(ca_descr->parity==0) { // even key
      set_even_control_word(keys[idx],ca_descr->cw);
      dc(printf("csa: even key set\n"))
      }
    else {                    // odd key
      set_odd_control_word(keys[idx],ca_descr->cw);
      dc(printf("csa: odd key set\n"))
      }
    }
  return true;
}

bool cDeCSA::SetCaPid(ca_pid_t *ca_pid)
{
  cMutexLock lock(this);
  if(ca_pid->index<MAX_CSA_IDX && ca_pid->pid<MAX_CSA_PIDS) {
    pidmap[ca_pid->pid]=ca_pid->index;
    dc(printf("csa: set pid %04x to index %d\n",ca_pid->pid,ca_pid->index))
    }
  return true;
}

bool cDeCSA::Decrypt(unsigned char *data, int len, bool force)
{
  cMutexLock lock(this);
  int r=-2, ccs=0, currIdx=-1;
  bool newRange=true;
  range[0]=0;
  len-=(TS_SIZE-1);
  for(int l=0; l<len; l+=TS_SIZE) {
    if(data[l]!=TS_SYNC_BYTE) {       // let higher level cope with that
      dc(printf("csa: garbage in TS buffer\n"))
      break;
      }
    if(data[l+3]&0xC0) {              // encrypted
      int idx=pidmap[((data[l+1]<<8)+data[l+2])&(MAX_CSA_PIDS-1)];
      if(currIdx<0 || idx==currIdx) { // same or no index
        currIdx=idx;
        if(newRange) {
          r+=2; newRange=false;
          range[r]=&data[l];
          range[r+2]=0;
          }
        range[r+1]=&data[l+TS_SIZE];
        if(++ccs>=cs) break;
        }
      else {                          // other index, create hole
        dc(printf("csa: other index. current=%d idx=%d\n",currIdx,idx))
        newRange=true;
        }
      }
    else {                            // unencrypted
      // nothing, we don't create holes for unencrypted packets
      }
    }
  if(force) dc(printf("###\ncsa: forced\n###\n"))
  if(r>cs*2) dc(printf("***\ncsa: range overflow\n***\n"))
  if(r<0) dc(printf("csa: nothing to decrypt\n"))
  if(r>=0 && (ccs>=cs || force) && GetKeyStruct(currIdx)) { // we have some range
    dc(printf("csa: decrypting %3d packets (ccs=%3d cs=%3d %s)",ccs,ccs,cs,ccs>=cs?"OK ":"INC"))
    int n=decrypt_packets(keys[currIdx],range);
    dc(printf(" -> %3d packets decrypted\n",n))
    if(n>0) return true;
    }
  else {
    if(ccs<cs && !force)
      dc(printf("csa: incomplete cluster ccs=%3d cs=%3d\n",ccs,cs))
    }
  return false;
}

// -- cDeCsaTSBuffer -----------------------------------------------------------

class cDeCsaTSBuffer : public cThread {
private:
  int f;
  int cardIndex;
  bool delivered;
  cRingBufferLinear *ringBuffer;
  //
  cDeCSA *decsa;
  unsigned char *lastP;
  int lastCount;
  //
  virtual void Action(void);
public:
  cDeCsaTSBuffer(int File, int Size, int CardIndex, cDeCSA *DeCsa);
  ~cDeCsaTSBuffer();
  uchar *Get(void);
  };

cDeCsaTSBuffer::cDeCsaTSBuffer(int File, int Size, int CardIndex, cDeCSA *DeCsa)
{
  SetDescription("TS buffer on device %d", CardIndex);
  f=File; cardIndex=CardIndex; decsa=DeCsa;
  delivered=false;
  lastP=0; lastCount=0;
  ringBuffer=new cRingBufferLinear(Size,TS_SIZE,true,"TS");
  ringBuffer->SetTimeouts(100,100);
  Start();
}

cDeCsaTSBuffer::~cDeCsaTSBuffer()
{
  Cancel(3);
  delete ringBuffer;
}

void cDeCsaTSBuffer::Action(void)
{
  if(ringBuffer) {
    bool firstRead=true;
    cPoller Poller(f);
    while(Running()) {
      if(firstRead || Poller.Poll(100)) {
        firstRead=false;
        int r=ringBuffer->Read(f);
        if(r<0 && FATALERRNO) {
          if(errno==EOVERFLOW)
            esyslog("ERROR: driver buffer overflow on device %d",cardIndex);
          else { LOG_ERROR; break; }
          }
        }
      }
    }
}

uchar *cDeCsaTSBuffer::Get(void)
{
  int Count=0;
  if(delivered) { ringBuffer->Del(TS_SIZE); delivered=false; }
  uchar *p=ringBuffer->Get(Count);
  if(p && Count>=TS_SIZE) {
    if(*p!=TS_SYNC_BYTE) {
      for(int i=1; i<Count; i++)
        if(p[i]==TS_SYNC_BYTE) { Count=i; break; }
      ringBuffer->Del(Count);
      esyslog("ERROR: skipped %d bytes to sync on TS packet on device %d",Count,cardIndex);
      return NULL;
      }

    if(decsa && (p[3]&0xC0)) {
      if(!decsa->Decrypt(p,Count,(lastP==p && lastCount==Count))) {
        lastP=p; lastCount=Count;
        cCondWait::SleepMs(20);
        return NULL;
        }
      lastP=0;
      }
    else p[3]&=~0xC0; // FF hack

    delivered=true;
    return p;
    }
  return NULL;
}

// -- cScDvbDevice -------------------------------------------------------------

cScDvbDevice::cScDvbDevice(int n, int cafd)
:cDvbDevice(n)
{
  decsa=0; tsBuffer=0; cam=0; ciadapter=0; hwciadapter=0;
  fd_ca=cafd; fd_ca2=dup(fd_ca); fd_dvr=-1;
  softcsa=(fd_ca<0);
}

cScDvbDevice::~cScDvbDevice()
{
  EarlyShutdown();
  delete ciadapter;
  delete hwciadapter;
  delete decsa;
  if(fd_ca>=0) close(fd_ca);
  if(fd_ca2>=0) close(fd_ca2);
}

void cScDvbDevice::EarlyShutdown(void)
{
  if(ciadapter) ciadapter->Shutdown();
  if(cam) cam->Stop();
  delete cam; cam=0;
}

void cScDvbDevice::LateInit(void)
{
  int n=CardIndex();
  if(softcsa) {
    if(HasDecoder()) LogError("Card %d is a full-featured card but no ca device found!",n);
    }
  else if(cSoftCAM::ForceBudget(n)) {
    LogInfo("Budget mode forced on card %d",n);
    softcsa=true;
    }
  
  if(fd_ca2>=0) hwciadapter=cDvbCiAdapter::CreateCiAdapter(this,fd_ca2);
  cam=new cCam(this,n);
  ciadapter=new cScCiAdapter(this,n,cam);
  if(softcsa) {
    LogInfo("Using software decryption on card %d",n);
    decsa=new cDeCSA;
    }
}

void cScDvbDevice::Shutdown(void)
{
  for(int n=cDevice::NumDevices(); --n>=0;) {
    cScDvbDevice *dev=dynamic_cast<cScDvbDevice *>(cDevice::GetDevice(n));
    if(dev) dev->EarlyShutdown();
    }
}

void cScDvbDevice::Startup(void)
{
  for(int n=cDevice::NumDevices(); --n>=0;) {
    cScDvbDevice *dev=dynamic_cast<cScDvbDevice *>(cDevice::GetDevice(n));
    if(dev) dev->LateInit();
    }
}

void cScDvbDevice::Capture(void)
{
  int i, found=0;
  for(i=0; i<MAXDVBDEVICES; i++) {
    if(UseDevice(NextCardIndex())) {
      char name[128];
      DvbName(DEV_DVB_FRONTEND,i,name,sizeof(name));
      if(access(name,F_OK)==0) {
        dsyslog("sc-plugin: probing %s",name);
        int f=open(name,O_RDONLY);
        if(f>=0) {
          close(f);
          dsyslog("sc-plugin: capturing device %d",i);
          new cScDvbDevice(i,DvbOpen(DEV_DVB_CA,i,O_RDWR));
          found++;
          }
        else {
          if(errno!=ENODEV && errno!=EINVAL) LOG_ERROR_STR(name);
          break;
          }
        }
      else {
        if(errno!=ENOENT) LOG_ERROR_STR(name);
        break;
        }
      }
    else NextCardIndex(1);
    }
  NextCardIndex(MAXDVBDEVICES-i);
  SetUseDevice(MAXDEVICES-1);
  if(found>0) isyslog("sc-plugin: captured %d video device%s",found,found>1 ? "s" : "");
  else isyslog("sc-plugin: no DVB device captured");
}

bool cScDvbDevice::Ready(void)
{
  return (ciadapter   ? ciadapter->Ready():true) &&
         (hwciadapter ? hwciadapter->Ready():true);
}

bool cScDvbDevice::SetPid(cPidHandle *Handle, int Type, bool On)
{
  if(cam) cam->SetPid(Type,Handle->pid,On);
  return cDvbDevice::SetPid(Handle,Type,On);
}

bool cScDvbDevice::SetChannelDevice(const cChannel *Channel, bool LiveView)
{
  if(cam) cam->Tune(Channel);
  return cDvbDevice::SetChannelDevice(Channel,LiveView);
}

bool cScDvbDevice::OpenDvr(void)
{
  CloseDvr();
  fd_dvr=DvbOpen(DEV_DVB_DVR,CardIndex(),O_RDONLY|O_NONBLOCK,true);
  if(fd_dvr>=0) tsBuffer=new cDeCsaTSBuffer(fd_dvr,MEGABYTE(2),CardIndex()+1,decsa);
  return fd_dvr>=0;
}

void cScDvbDevice::CloseDvr(void)
{
  if(fd_dvr>=0) {
    delete tsBuffer; tsBuffer=NULL;
    close(fd_dvr); fd_dvr=-1;
    }
}

bool cScDvbDevice::GetTSPacket(uchar *&Data)
{
  if(tsBuffer) { Data=tsBuffer->Get(); return true; }
  return false;
}

bool cScDvbDevice::SetCaDescr(ca_descr_t *ca_descr)
{
  if(!softcsa) {
    cMutexLock lock(&cafdMutex);
    return ioctl(fd_ca,CA_SET_DESCR,ca_descr)>=0;
    }
  else if(decsa) return decsa->SetDescr(ca_descr);
  return false;
}

bool cScDvbDevice::SetCaPid(ca_pid_t *ca_pid)
{
  if(!softcsa) {
    cMutexLock lock(&cafdMutex);
    return ioctl(fd_ca,CA_SET_PID,ca_pid)>=0;
    }
  else if(decsa) return decsa->SetCaPid(ca_pid);
  return false;
}

#ifdef DEBUG_AV7110
static unsigned int av7110_read(int fd, unsigned int addr)
{
  ca_pid_t arg;
  arg.pid=addr;
  ioctl(fd,CA_GET_MSG,&arg);
  return arg.index;
}

#if 0
static void av7110_write(int fd, unsigned int addr, unsigned int val)
{
  ca_pid_t arg;
  arg.pid=addr;
  arg.index=val;
  ioctl(fd,CA_SEND_MSG,&arg);
}
#endif
#endif

void cScDvbDevice::DumpAV7110(void)
{
#ifdef DEBUG_AV7110
#define CODEBASE (0x2e000404+0x1ce00)
  cMutexLock lock(&cafdMutex);
  if(HasDecoder() && lastDump.Elapsed()>20000) {
    lastDump.Set();
    static unsigned int handles=0, hw_handles=0;
    static const unsigned int code[] = {
      0xb5100040,0x4a095a12,0x48094282,0xd00b4b09,0x20000044,
      0x5b1c4294,0xd0033001,0x281cdbf8,0xe001f7fe,0xfd14bc10
      };
    static const unsigned int mask[] = {
      0xffffffff,0xffffffff,0xffffffff,0xffffffff,0xffffffff,
      0xffffffff,0xffffffff,0xffffffff,0xfffff800,0xf800ffff
      };
    if(!handles) {
      handles=1;
      printf("searching handle tables\n");
      for(int i=0; i<0x2000; i+=4) {
        int j;
        for(j=0; j<20; j+=4) {
          int r=av7110_read(fd_ca,CODEBASE+i+j);
          if((r&mask[j/4])!=(code[j/4]&mask[j/4])) break;
          }
        if(j==20) {
          handles=av7110_read(fd_ca,CODEBASE+i+44);
          hw_handles=av7110_read(fd_ca,CODEBASE+i+52);
          printf("found handles=%08x hw_handles=%08x at 0x%08x\n",handles,hw_handles,CODEBASE+i);
          if((handles>>16)!=0x2e08 || (hw_handles>>16)!=0x2e08) {
            printf("seems to be invalid\n");
            }
          break;
          }
        }
      }

    unsigned int hdl=0, hwhdl=0;
    printf("         : 64000080 64000400\n");
    for(int i=0; i<=31; i++) {
      unsigned int off80 =av7110_read(fd_ca,0x64000080+i*4);
      unsigned int off400=av7110_read(fd_ca,0x64000400+i*8);
      printf("handle %2d: %08x %08x %s pid=%04x idx=%d",
        i,off80,off400,off80&0x2000?"ACT":"---",off80&0x1fff,(off400&0x1e)>>1);
      if(handles>1 && i<=27) {
        if((i&1)==0) {
          hdl=av7110_read(fd_ca,handles+i*2);
          hwhdl=av7110_read(fd_ca,hw_handles+i*2);
          }
        unsigned int s=((~i)&1)<<4;
        printf(" | %02d hdl=%04x hwfilt=%04x",i,(hdl>>s)&0xffff,(hwhdl>>s)&0xffff);
        }
      printf("\n");
      }
    }
#endif
}
