/*
 * Softcam plugin to VDR (C++)
 *
 * This code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <unistd.h>
#include <time.h>

#include "log.h"
#include "misc.h"

static const struct LogModule lm_general = {
  (LMOD_ENABLE|L_GEN_ERROR|L_GEN_WARN|L_GEN_INFO|L_GEN_DEBUG)&LOPT_MASK,
  (LMOD_ENABLE|L_GEN_ERROR|L_GEN_WARN|L_GEN_INFO|L_GEN_DEBUG)&LOPT_MASK,
  "general", { "error","warn","info","debug" }
  };
ADD_MODULE(L_GEN,lm_general)

// -- cLogging -----------------------------------------------------------------

#define LMOD_SUP 32
#define LMOD_CFG_VALID  0x80000000

struct LogHeader {
  char stamp[32];
  char tag[64];
  };

static const struct LogModule *mods[LMOD_SUP];
static int config[LMOD_SUP];
static int logmode=LOG_CON;

bool cLogging::AddModule(int m, const struct LogModule *lm)
{
  if(m<LMOD_SUP) {
    mods[m]=lm;
    if(config[m]&LMOD_CFG_VALID) config[m]&=(lm->OptSupported&LOPT_MASK);
    else SetModuleDefault(LCLASS(m,0));
//XXX
printf("module '%s' added. mod=%x supp=%x def=%x cfg=%x\n",lm->Name,m,lm->OptSupported,lm->OptDefault,config[m]);
    }
  else Printf(L_GEN_ERROR,"failed to add logging module %d (%s)",m,lm->Name);
  return true;
}

const struct LogModule *cLogging::GetModule(int c)
{
  int m=LMOD(c);
  return m<LMOD_SUP ? mods[m] : 0;
}

bool cLogging::GetHeader(int c, struct LogHeader *lh)
{
  const struct LogModule *lm=GetModule(c);
  if(lm) {
    time_t tt=time(0);
    struct tm tm_r;
    strftime(lh->stamp,sizeof(lh->stamp),"%b %e %T",localtime_r(&tt,&tm_r));
    int i, o=LOPT(c);
    for(i=0; i<32; i++,o>>=1) if(o&1) break;
    snprintf(lh->tag,sizeof(lh->tag),"%s.%s",lm->Name,lm->OptName[i]);
    return true;
    }
  return false;  
}

void cLogging::LogLine(const struct LogHeader *lh, const char *txt)
{
  if(logmode&LOG_CON)
    printf("%s %s %s\n",lh->stamp,lh->tag,txt);

  if(logmode&LOG_FILE) {
    }

  if(logmode&LOG_SYS) {
    }
}

bool cLogging::Enabled(int c)
{
  int m=LMOD(c);
  int o=LOPT(c)|LMOD_CFG_VALID|LMOD_ENABLE;
  return m<LMOD_SUP && (config[m] & o)==o;
}

void cLogging::Printf(int c, const char *format, ...)
{
  if(Enabled(c)) {
    struct LogHeader lh;
    if(GetHeader(c,&lh)) {
      char buff[1024];
      va_list ap;
      va_start(ap,format);
      vsnprintf(buff,sizeof(buff),format,ap);
      va_end(ap);
      LogLine(&lh,buff);
      }
    }
}

void cLogging::Puts(int c, const char *txt)
{
  if(txt && Enabled(c)) {
    struct LogHeader lh;
    if(GetHeader(c,&lh)) {
      LogLine(&lh,txt);
      }
    }
}

void cLogging::PutLB(int c, cLineBuff *lb)
{
  const char *txt=lb->Line();
  Puts(c,txt);
  lb->Flush();
}

void cLogging::Dump(int c, const unsigned char *data, int n, const char *format, ...)
{
  if(Enabled(c)) {
    struct LogHeader lh;
    if(GetHeader(c,&lh)) {
      char buff[1024];
      va_list ap;
      va_start(ap,format);
      vsnprintf(buff,sizeof(buff),format,ap);
      va_end(ap);
      LogLine(&lh,buff);
      for(int i=0; i<n;) {
        int q=sprintf(buff,"%04x:",i);
        for(int l=0 ; l<16 && i<n ; l++) q+=sprintf(&buff[q]," %02x",data[i++]);
        LogLine(&lh,buff);
        }
      }
    }
}

void cLogging::SetModuleDefault(int c)
{
  const struct LogModule *lm=GetModule(c);
  if(lm) config[LMOD(c)]=(lm->OptDefault&LOPT_MASK)|LMOD_CFG_VALID;
}
