/*
 * Softcam plugin to VDR (C++)
 *
 * This code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html
 */

#include <malloc.h>
#include <stdlib.h>
#include <getopt.h>
#include <typeinfo>
#ifndef STATICBUILD
#include <dlfcn.h>
#include <dirent.h>
#include <fnmatch.h>
#endif

#include <vdr/plugin.h>
#include <vdr/menuitems.h>
#include <vdr/status.h>
#include <vdr/dvbdevice.h>
#include <vdr/channels.h>
#include <vdr/ci.h>
#include <vdr/interface.h>
#include <vdr/menu.h>
#include <vdr/tools.h>
#include <vdr/config.h>

#include "common.h"
#include "sc.h"
#include "scsetup.h"
#include "filter.h"
#include "system.h"
#include "cam.h"
#include "smartcard.h"
#include "data.h"
#include "network.h"
#include "misc.h"
#include "opts.h"
#include "i18n.h"

#define MIN_VERS   1 // required VDR version
#define MIN_MAJOR  5
#define MIN_MINOR  0
#define MINAPIVERSNUM 10500

#define DEBUG_TRCAIDS

// some sanity checks
#ifdef HAVE_SOFTCSA
#error softcsa/ffdecsa patch MUST NOT be applied. Next time read the README first.
#endif
#ifdef VDR_IS_SC_PATCHED
#error You MUST NOT patch the VDR core. Next time read the README first.
#endif
#if APIVERSNUM<MINAPIVERSNUM
#error Your VDR API version is too old. See README.
#endif

// SC API version number for loading shared libraries
#define SCAPIVERS 6

static const char *VERSION        = "0.7.0";
static const char *DESCRIPTION    = "A software emulated CAM";

cPlugin *ScPlugin;

#ifdef DEBUG_TRCAIDS
#define dt(x) { x; }
#else
#define dt(x) ;
#endif

// --- cMenuEditCapItem --------------------------------------------------------

class cMenuEditCapItem : public cMenuEditIntItem {
protected:
  virtual void Set(void);
public:
  cMenuEditCapItem(const char *Name, int *Value);
  eOSState ProcessKey(eKeys Key);
  };

cMenuEditCapItem::cMenuEditCapItem(const char *Name, int *Value)
:cMenuEditIntItem(Name, Value, 0)
{
  Set();
}

void cMenuEditCapItem::Set(void)
{
  if(!*value) SetValue(tr("off"));
  else cMenuEditIntItem::Set();
}

eOSState cMenuEditCapItem::ProcessKey(eKeys Key)
{
  eOSState state = cMenuEditItem::ProcessKey(Key);
  if(state==osUnknown)
    state=cMenuEditIntItem::ProcessKey(Key);
  return state;
}

// --- cMenuEditHexItem ------------------------------------------------------

class cMenuEditHexItem : public cMenuEditItem {
private:
  bool abc, isOn;
  //
  void SetButtons(bool on);
protected:
  int *value;
  int min, max;
  //
  virtual void Set(void);
public:
  cMenuEditHexItem(const char *Name, int *Value, int Min=0, int Max=INT_MAX);
  virtual eOSState ProcessKey(eKeys Key);
  };

cMenuEditHexItem::cMenuEditHexItem(const char *Name, int *Value, int Min, int Max)
:cMenuEditItem(Name)
{
  value=Value; min=Min; max=Max;
  if(*value<min) *value=min;
  else if(*value>max) *value=max;
  Set();
  abc=true; isOn=false;
}

void cMenuEditHexItem::SetButtons(bool on)
{
  if(on) {
    if(abc) cSkinDisplay::Current()->SetButtons("A","B","C","D-F");
    else    cSkinDisplay::Current()->SetButtons("D","E","F","A-C");
    isOn=true;
    }
  else {
    cSkinDisplay::Current()->SetButtons(0);
    isOn=false;
    }
}

void cMenuEditHexItem::Set(void)
{
   char buf[16];
   snprintf(buf,sizeof(buf),"%X",*value);
   SetValue(buf);
}

eOSState cMenuEditHexItem::ProcessKey(eKeys Key)
{
  switch(NORMALKEY(Key)) {
    case kUp:
    case kDown:
      if(isOn) SetButtons(false);
      break;
    default:
      if(!isOn) SetButtons(true);
      break;
    }
  eOSState state=cMenuEditItem::ProcessKey(Key);
  if(state!=osUnknown) return state;

  int newValue=*value;
  bool IsRepeat=Key & k_Repeat;
  Key=NORMALKEY(Key);
  switch(Key) {
    case kBlue:
      abc=!abc; SetButtons(true);
      break;
    case kRed:
    case kGreen:
    case kYellow:
    case k0 ... k9:
      {
      if(fresh) { newValue=0; fresh=false; }
      int add;
      if(Key>=kRed && Key<=kYellow) add=(abc ? 10:13)+(Key-kRed);
      else                          add=(Key-k0);
      newValue=newValue*16+add;
      break;
      }
    case kLeft:
      newValue=*value-1; fresh=true;
      if(!IsRepeat && newValue<min) newValue=max;
      break;
    case kRight:
      newValue=*value+1; fresh=true;
      if(!IsRepeat && newValue>max) newValue=min;
      break;
    default:
      if(*value<min) { *value=min; Set(); }
      if(*value>max) { *value=max; Set(); }
      return osUnknown;
    }
  if(newValue!=*value && (!fresh || min<=newValue) && newValue<=max) {
     *value=newValue;
     Set();
     }
  return osContinue;
}

// --- cScInfoItem -------------------------------------------------------------

class cScInfoItem : public cOsdItem {
private:
  void SetValue(const char *Name, const char *Value);
public:
  cScInfoItem(const char *Name, int Value, eOSState State=osUnknown);
  cScInfoItem(const char *Name, const char *Value=0, eOSState State=osUnknown);
  };

cScInfoItem::cScInfoItem(const char *Name, int Value, eOSState State)
:cOsdItem(State)
{
  char buf[16];
  snprintf(buf,sizeof(buf),"%d",Value);
  SetValue(Name,buf);
  if(State==osUnknown) SetSelectable(false);
}

cScInfoItem::cScInfoItem(const char *Name, const char *Value, eOSState State)
:cOsdItem(State)
{
  SetValue(Name,Value);
  if(State==osUnknown) SetSelectable(false);
}

void cScInfoItem::SetValue(const char *Name, const char *Value)
{
  char *buff;
  asprintf(&buff,Value ? "%s:\t%s":"%s",Name,Value);
  SetText(buff,false);
  cStatus::MsgOsdCurrentItem(buff);
}

// --- cOpt --------------------------------------------------------------------

cOpt::cOpt(const char *Name, const char *Title)
{
  name=Name; title=Title;
  fullname=0; persistant=true;
}

cOpt::~cOpt()
{
  free(fullname);
}

const char *cOpt::FullName(const char *PreStr)
{
  if(PreStr) {
    free(fullname);
    asprintf(&fullname,"%s.%s",PreStr,name);
    return fullname;
    }
  else return name;
}

// --- cOptInt -----------------------------------------------------------------

cOptInt::cOptInt(const char *Name, const char *Title, int *Storage, int Min, int Max)
:cOpt(Name,Title)
{
  storage=Storage; min=Min; max=Max;
}

void cOptInt::Parse(const char *Value)
{
  *storage=atoi(Value);
}

void cOptInt::Backup(void)
{
  value=*storage;
}

bool cOptInt::Set(void)
{
  if(value!=*storage) { *storage=value; return true; }
  return false;
}

void cOptInt::Store(const char *PreStr)
{
  ScPlugin->SetupStore(FullName(PreStr),*storage);
}

void cOptInt::Create(cOsdMenu *menu)
{
  menu->Add(new cMenuEditIntItem(tr(title),&value,min,max));
}

// --- cOptSel -----------------------------------------------------------------

cOptSel::cOptSel(const char *Name, const char *Title, int *Storage, int NumStr, const char * const *Strings)
:cOptInt(Name,Title,Storage,0,NumStr)
{
  strings=Strings;
  trStrings=0;
}

cOptSel::~cOptSel()
{
  free(trStrings);
}

void cOptSel::Create(cOsdMenu *menu)
{
  free(trStrings);
  if((trStrings=MALLOC(const char *,max))) {
    for(int i=0; i<max ; i++) trStrings[i]=tr(strings[i]);
    menu->Add(new cMenuEditStraItem(tr(title),&value,max,trStrings));
    }
}

// --- cOptBool -----------------------------------------------------------------

cOptBool::cOptBool(const char *Name, const char *Title, int *Storage)
:cOptInt(Name,Title,Storage,0,1)
{}

void cOptBool::Create(cOsdMenu *menu)
{
  menu->Add(new cMenuEditBoolItem(tr(title),&value));
}

// --- cOptStr -----------------------------------------------------------------

cOptStr::cOptStr(const char *Name, const char *Title, char *Storage, int Size, const char *Allowed)
:cOpt(Name,Title)
{
  storage=Storage; size=Size; allowed=Allowed;
  value=MALLOC(char,size);
}

cOptStr::~cOptStr()
{
  free(value);
}

void cOptStr::Parse(const char *Value)
{
  strn0cpy(storage,Value,size);
}

void cOptStr::Backup(void)
{
  strn0cpy(value,storage,size);
}

bool cOptStr::Set(void)
{
  if(strcmp(value,storage)) { strn0cpy(storage,value,size); return true; }
  return false;
}

void cOptStr::Store(const char *PreStr)
{
  ScPlugin->SetupStore(FullName(PreStr),storage);
}

void cOptStr::Create(cOsdMenu *menu)
{
  menu->Add(new cMenuEditStrItem(tr(title),value,size,allowed));
}

// --- cOptMInt ----------------------------------------------------------------

class cOptMInt : public cOpt {
protected:
  int *storage, *value;
  int size, mode, len;
public:
  cOptMInt(const char *Name, const char *Title, int *Storage, int Size, int Mode);
  virtual ~cOptMInt();
  virtual void Parse(const char *Value);
  virtual void Backup(void);
  virtual bool Set(void);
  virtual void Store(const char *PreStr);
  virtual void Create(cOsdMenu *menu);
  };

// mode: 0-Cap 1-Int 2-Hex

cOptMInt::cOptMInt(const char *Name, const char *Title, int *Storage, int Size, int Mode)
:cOpt(Name,Title)
{
  storage=Storage; size=Size; mode=Mode; len=sizeof(int)*size;
  value=MALLOC(int,size);
}

cOptMInt::~cOptMInt()
{
  free(value);
}

void cOptMInt::Parse(const char *Value)
{
  memset(storage,0,len);
  int i=0;
  while(1) {
    char *p;
    const int c=strtol(Value,&p,mode>1 ? 16:10);
    if(p==Value || i>=size) return;
    if(c>0) storage[i++]=c;
    Value=p;
    }
}

void cOptMInt::Backup(void)
{
  memcpy(value,storage,len);
}

bool cOptMInt::Set(void)
{
  if(memcmp(value,storage,len)) {
    memset(storage,0,len);
    for(int i=0, k=0; i<size; i++) if(value[i]>0) storage[k++]=value[i];
    return true;
    }
  return false;
}

void cOptMInt::Store(const char *PreStr)
{
  char b[256];
  int p=0;
  for(int i=0; i<size; i++)
    if(storage[i]) p+=snprintf(b+p,sizeof(b)-p,mode>1 ? "%x ":"%d ",storage[i]);
  ScPlugin->SetupStore(FullName(PreStr),p>0?b:0);
}

void cOptMInt::Create(cOsdMenu *menu)
{
  for(int i=0; i<size; i++) {
    const char *buff=tr(title);
    switch(mode) {
      case 0: menu->Add(new cMenuEditCapItem(buff,&value[i])); break;
      case 1: menu->Add(new cMenuEditIntItem(buff,&value[i],0,65535)); break;
      case 2: menu->Add(new cMenuEditHexItem(buff,&value[i],0,65535)); break;
      }
    if(value[i]==0) break;
    }
}

// --- cOpts -------------------------------------------------------------------

cOpts::cOpts(const char *PreStr, int NumOpts)
{
  preStr=PreStr;
  numOpts=NumOpts; numAdd=0;
  if((opts=MALLOC(cOpt *,numOpts))) memset(opts,0,sizeof(cOpt *)*numOpts);
}

cOpts::~cOpts()
{
  if(opts) {
    for(int i=0; i<numOpts; i++) delete opts[i];
    free(opts);
    }
}

void cOpts::Add(cOpt *opt)
{
  if(opts && numAdd<numOpts) opts[numAdd++]=opt;
}

bool cOpts::Parse(const char *Name, const char *Value)
{
  if(opts) {
    for(int i=0; i<numAdd; i++)
      if(opts[i] && opts[i]->Persistant() && !strcasecmp(Name,opts[i]->Name())) {
        opts[i]->Parse(Value);
        return true;
        }
    }
  return false;
}

bool cOpts::Store(bool AsIs)
{
  bool res=false;
  if(opts) {
    for(int i=0; i<numAdd; i++)
      if(opts[i]) {
        if(!AsIs && opts[i]->Set()) res=true;
        if(opts[i]->Persistant()) opts[i]->Store(preStr);
        }
    }
  return res;
}

void cOpts::Backup(void)
{
  if(opts) {
    for(int i=0; i<numAdd; i++)
      if(opts[i]) opts[i]->Backup();
    }
}

void cOpts::Create(cOsdMenu *menu)
{
  if(opts) {
    for(int i=0; i<numAdd; i++)
      if(opts[i]) opts[i]->Create(menu);
    }
}

// --- cMenuInfoSc -------------------------------------------------------------

class cMenuInfoSc : public cOsdMenu {
public:
  cMenuInfoSc(void);
  virtual eOSState ProcessKey(eKeys Key);
  };

cMenuInfoSc::cMenuInfoSc(void)
:cOsdMenu(tr("SoftCAM"),25)
{
  Add(new cScInfoItem(tr("Current keys:")));
  for(int d=0; d<MAXDVBDEVICES; d++) {
    int n=0;
    char *ks;
    do {
      if((ks=cSoftCAM::CurrKeyStr(d,n))) {
        char buffer[32];
        snprintf(buffer,sizeof(buffer),"  [%s %d:%d]","DVB",d+1,n+1);
        Add(new cScInfoItem(buffer,ks));
        free(ks);
        }
      n++;
      } while(ks);
    }
  if(Feature.KeyFile()) {
    Add(new cScInfoItem(tr("Key update status:")));
    int fk, nk;
    cSoftCAM::KeyStats(fk,nk);
    Add(new cScInfoItem(tr("  [Seen keys]"),fk));
    Add(new cScInfoItem(tr("  [New keys]"), nk));
    }
  Display();
}

eOSState cMenuInfoSc::ProcessKey(eKeys Key)
{
  eOSState state=cOsdMenu::ProcessKey(Key);
  if(state==osUnknown && Key==kOk) state=osBack;
  return state;
}

// --- cMenuInfoCard -----------------------------------------------------------

class cMenuInfoCard : public cMenuText {
private:
  int port;
  char infoStr[1024];
public:
  cMenuInfoCard(int Port);
  virtual eOSState ProcessKey(eKeys Key);
  };

cMenuInfoCard::cMenuInfoCard(int Port)
:cMenuText(tr("Smartcard"),0,fontFix)
{
  port=Port;
  smartcards.CardInfo(port,infoStr,sizeof(infoStr));
  SetText(infoStr);
  SetHelp(tr("Reset card"));
  Display();
}

eOSState cMenuInfoCard::ProcessKey(eKeys Key)
{
  if(Key==kRed && Interface->Confirm(tr("Really reset card?"))) {
    smartcards.CardReset(port);
    return osEnd;
    }
  eOSState state=cMenuText::ProcessKey(Key);
  if(state==osUnknown) state=osContinue;
  return state;
}

// --- cMenuSetupSc ------------------------------------------------------------

static cOpts *ScOpts;

class cMenuSetupSc : public cMenuSetupPage {
private:
  char *cfgdir;
  //
  void SystemOpts(void);
protected:
  virtual void Store(void);
public:
  cMenuSetupSc(const char *CfgDir);
  virtual ~cMenuSetupSc();
  virtual eOSState ProcessKey(eKeys Key);
  };

static eOSState portStates[] = { osUser1,osUser2,osUser3,osUser4 };
#if MAX_PORTS!=4
#error Update portStates[]
#endif

cMenuSetupSc::cMenuSetupSc(const char *CfgDir)
{
  cfgdir=strdup(CfgDir);
  SetSection(tr("SoftCAM"));
  ScOpts->Backup(); ScOpts->Create(this);
  if(Feature.SmartCard()) {
    char id[IDSTR_LEN];
    for(int i=0; smartcards.ListCard(i,id,sizeof(id)); i++) {
      char buff[32];
      snprintf(buff,sizeof(buff),"%s %d",tr("Smartcard interface"),i);
      if(id[0]) Add(new cScInfoItem(buff,id,portStates[i]));
      else      Add(new cScInfoItem(buff,tr("(empty)")));
      }
    }
  SystemOpts();
  Add(new cOsdItem(tr("Status information..."),osUser8));
  Add(new cOsdItem(tr("Flush ECM cache"),osUser7));
  Add(new cOsdItem(tr("Reload files"),osUser9));
}

cMenuSetupSc::~cMenuSetupSc()
{
  free(cfgdir);
}

void cMenuSetupSc::SystemOpts(void)
{
  bool start=true;
  cOpts *opts;
  while((opts=cSystems::GetSystemOpts(start))) {
    opts->Backup();
    opts->Create(this);
    start=false;
    }
}

void cMenuSetupSc::Store(void)
{
  cSoftCAM::StoreConfig(false);
}

eOSState cMenuSetupSc::ProcessKey(eKeys Key)
{
  eOSState state = cOsdMenu::ProcessKey(Key);

  switch(state) {
    case osUser1...osUser6:
      if(Feature.SmartCard()) {
        for(unsigned int i=0; i<sizeof(portStates)/sizeof(eOSState); i++)
          if(portStates[i]==state) return(AddSubMenu(new cMenuInfoCard(i)));
        }
      state=osContinue;
      break;

    case osUser7:
      if(Interface->Confirm(tr("Really flush ECM cache?"))) {
        ecmcache.Flush();
        cSoftCAM::Save();
        state=osEnd;
        }
      break;

    case osUser8:
      return(AddSubMenu(new cMenuInfoSc));

    case osUser9:
      state=osContinue;
      if(!cSoftCAM::Active()) {
        if(Interface->Confirm(tr("Really reload files?"))) {
          Store();
          cSoftCAM::Load(cfgdir);
          state=osEnd;
          }
        }
      else 
        Skins.Message(mtError,tr("Active! Can't reload files now"));
      break;

    case osContinue:
      if(NORMALKEY(Key)==kUp || NORMALKEY(Key)==kDown) {
        cOsdItem *item=Get(Current());
        if(item) item->ProcessKey(kNone);
        }
      break;

    case osUnknown:
      if(Key==kOk) { Store(); state=osBack; }
      break;

    default:
      break;
    }
  return state;
}

// --- cScSetup ---------------------------------------------------------------

cScSetup ScSetup;

cScSetup::cScSetup(void)
{
  AutoUpdate = 1;
  memset(ScCaps,0,sizeof(ScCaps));
  ScCaps[0] = 1;
  ScCaps[1] = 2;
  ConcurrentFF = 0;
  memset(CaIgnore,0,sizeof(CaIgnore));
  LocalPriority = 0;
}

void cScSetup::CheckConfig(void)
{
  if(AutoUpdate==0)
    LogWarn("Keys updates (AU) are disabled.");
  for(int i=0; i<MAXSCCAPS; i++)
    if(ScCaps[i]>=16) {
      LogWarn("ScCaps contains unusual value. Check your config! (You can ignore this message if you have more than 16 dvb cards in your system ;)");
      break;
      }

#ifdef DEBUG
  printf("** Plugin config:\n");
  printf("** Key updates (AU) are %s\n",AutoUpdate?(AutoUpdate==1?"enabled (active CAIDs)":"enabled (all CAIDs)"):"DISABLED");
  printf("** Local systems %stake priority over cached remote\n",LocalPriority?"":"DONT'T ");
  printf("** ScCaps are"); for(int i=0; i<MAXSCCAPS ; i++) printf(" %d",ScCaps[i]); printf("\n");
  printf("** Concurrent FF recordings are %sallowed\n",ConcurrentFF?"":"NOT ");
  printf("** Ignored CAIDs"); for(int i=0; i<MAXCAIGN ; i++) printf(" %04X",CaIgnore[i]); printf("\n");
#endif
}

// --- cTransponderCaids -------------------------------------------------------

class cTransponderCaids : public cSimpleItem {
private:
  int source, transponder;
  int numcaids;
  int caids[MAX_CI_SLOT_CAIDS+1];
public:
  cTransponderCaids(int Source, int Transponder);
  void AddCaids(const int *ids);
  void Del(int caid);
  void Sort(void);
  int NumDiff(cTransponderCaids *tr);
  bool IsTransponder(int Source, int Transponder);
  void Dump(void);
  const int *Caids(void);
  bool CanAdd(int n) { return numcaids+n<=MAX_CI_SLOT_CAIDS; }
  };

cTransponderCaids::cTransponderCaids(int Source, int Transponder)
{
  source=Source; transponder=Transponder;
  numcaids=0;
}

bool cTransponderCaids::IsTransponder(int Source, int Transponder)
{
  return source==Source && transponder==Transponder;
}

const int *cTransponderCaids::Caids(void)
{
  caids[numcaids]=0;
  return caids;
}

void cTransponderCaids::AddCaids(const int *ids)
{
  for(; *ids; ids++) {
    int i;
    for(i=0; i<numcaids; i++) if(caids[i]==*ids) break;
    if(i==numcaids && numcaids<MAX_CI_SLOT_CAIDS) caids[numcaids++]=*ids;
    }
}

void cTransponderCaids::Del(int caid)
{
  for(int i=0; i<numcaids; i++)
    if(caids[i]==caid) {
      numcaids--;
      caids[i]=caids[numcaids];
      caids[numcaids]=0;
      break;
      }
}

void cTransponderCaids::Sort(void)
{
  int tmp[MAX_CI_SLOT_CAIDS];
  int c=0xFFFF;
  for(int i=0; i<numcaids; i++) {
    int d=0;
    for(int j=0; j<numcaids; j++) if(caids[j]>d && caids[j]<c) d=caids[j];
    tmp[i]=d; c=d;
    }
  memcpy(caids,tmp,sizeof(caids));
}

int cTransponderCaids::NumDiff(cTransponderCaids *tr)
{
  int diff=0;
  for(int i=tr->numcaids-1; i>=0; i--) {
    int j;
    for(j=numcaids-1; j>=0; j--) if(caids[j]==tr->caids[i]) break;
    if(j<0) diff++;
    }
  return diff;
}

void cTransponderCaids::Dump(void)
{
#ifdef DEBUG_TRCAIDS
  printf("transponder %x/%x",source,transponder);
  const int *ids=Caids();
  for(; *ids; ids++) printf(" %04x",*ids);
  printf("\n");
#endif
}

// --- cTransponderList -------------------------------------------------------

class cTransponderList : public cSimpleList<cTransponderCaids> {
private:
public:
  cTransponderCaids *FindTransponder(int Source, int Transponder);
  void Sort(void);
  void Diff(int max);
  void CheckIgnore(void);
  };

cTransponderCaids *cTransponderList::FindTransponder(int Source, int Transponder)
{
  cTransponderCaids *tr;
  for(tr=First(); tr; tr=Next(tr)) if(tr->IsTransponder(Source,Transponder)) break;
  return tr;
}

void cTransponderList::Sort(void)
{
  for(cTransponderCaids *tr=First(); tr; tr=Next(tr))
    tr->Sort();
#ifdef DEBUG_TRCAIDS
  printf("after sort\n");
  for(cTransponderCaids *tr=First(); tr; tr=Next(tr)) tr->Dump();
#endif
}

void cTransponderList::CheckIgnore(void)
{
  for(cTransponderCaids *tr=First(); tr; tr=Next(tr)) {
    const int *ids=tr->Caids();
    while(*ids) {
      int pri=0;
      if(!cSystems::FindIdentBySysId(*ids,false,pri)) tr->Del(*ids);
      else ids++;
      }
    tr->Sort();
    }
#ifdef DEBUG_TRCAIDS
  printf("after check\n");
  for(cTransponderCaids *tr=First(); tr; tr=Next(tr)) tr->Dump();
#endif
}

void cTransponderList::Diff(int max)
{
  for(cTransponderCaids *tr1=First(); tr1; tr1=Next(tr1)) {
    for(cTransponderCaids *tr2=Next(tr1); tr2;) {
      int diff;
      if((diff=tr1->NumDiff(tr2))<=max && tr1->CanAdd(diff)) {
        if(diff>0) {
          tr1->AddCaids(tr2->Caids());
          tr1->Sort();
          }
        cTransponderCaids *t=Next(tr2);
        Del(tr2);
        tr2=t;
        }
      else tr2=Next(tr2);
      }
    }
#ifdef DEBUG_TRCAIDS
  printf("after diff %d\n",max);
  for(cTransponderCaids *tr=First(); tr; tr=Next(tr)) tr->Dump();
#endif
}

// --- cSoftCAM ---------------------------------------------------------------

#define CAID_TIME 300000 // time between caid scans

struct Caids {
  int version;
  int caids[MAX_CI_SLOT_CAIDS+1];
  };

static struct Caids caids[MAX_CI_SLOTS];
static cTimeMs caidTimer;
static cMutex caidMutex;

int cSoftCAM::foundKeys=0;
int cSoftCAM::newKeys=0;
int cSoftCAM::budget=0;

bool cSoftCAM::Load(const char *cfgdir)
{
  ecmcache.Load();
  if(Feature.KeyFile() && !keys.Load(cfgdir)) LogError("no keys loaded for softcam!");
  if(!cSystems::Init(cfgdir)) return false;
  if(Feature.SmartCard()) smartcards.LoadData(cfgdir);
  cLoaders::LoadCache(cfgdir);
  Save();
  PredictCaids(true);
  return true;
}

void cSoftCAM::Save(void)
{
  cLoaders::SaveCache();
}

void cSoftCAM::StoreConfig(bool AsIs)
{
  if(ScOpts) ScOpts->Store(AsIs);
  cSystems::ConfigStore(AsIs);
}

void cSoftCAM::ShutDown(void)
{
  LogStatsDown();
  cSystems::Clean();
}

bool cSoftCAM::CapCheck(int n)
{
  for(int j=0; j<MAXSCCAPS; j++)
    if(ScSetup.ScCaps[j] && ScSetup.ScCaps[j]==n+1) return true;
  return false;
}

int cSoftCAM::GetCaids(int slot, unsigned short *Caids, int max)
{
  cMutexLock lock(&caidMutex);
  struct Caids *ca=&caids[slot];
  if(Caids) {
    int i;
    for(i=0; i<MAX_CI_SLOT_CAIDS && i<max && ca->caids[i]; i++) Caids[i]=ca->caids[i];
    Caids[i]=0;
    }
  return ca->version;
}

void cSoftCAM::PredictCaids(bool force)
{
  if(caidTimer.TimedOut() || force) {
    dt(printf("softcam: building caid lists\n"))
    cTransponderList list;
    Channels.Lock(false);
    for(cChannel *channel=Channels.First(); channel; channel=Channels.Next(channel)) {
      if(!channel->GroupSep() && channel->Ca()>=CA_ENCRYPTED_MIN) {
        cTransponderCaids *tr=list.FindTransponder(channel->Source(),channel->Transponder());
        if(!tr) {
          tr=new cTransponderCaids(channel->Source(),channel->Transponder());
          if(tr) list.Add(tr);
          }
        if(tr) tr->AddCaids(channel->Caids());
        }
      }
    Channels.Unlock();
    list.Sort();
    list.Diff(0);
    list.CheckIgnore();
    int num;
    for(int max=1; max<=4; max++)
      do {
        num=list.Count();
        list.Diff(max);
        } while(list.Count()!=num);

    if(list.Count()>MAX_CI_SLOTS)
      d(printf("softcam: too many CAIDs. Would need more than %d CI slots. You should ignore some CAIDs.\n",MAX_CI_SLOTS))

    cMutexLock lock(&caidMutex);
    num=0;
    for(cTransponderCaids *tr=list.First(); num<MAX_CI_SLOTS && tr; num++, tr=list.Next(tr)) {
      struct Caids *ca=&caids[num];
      const int *ids=tr->Caids();
      int i;
      for(i=0; ids[i]; i++) if(ids[i]!=ca->caids[i]) break;
      if(ids[i] || ca->caids[i]) {
        ca->version++;
        i=0; do { ca->caids[i]=ids[i]; } while(ids[i++]);
        }
      }

#ifdef DEBUG
    printf("softcam: active caids\n");
    for(int j=0; j<MAX_CI_SLOTS; j++) {
      struct Caids *ca=&caids[j];
      if(ca->version>0) {
        printf("slot %d:",j);
        for(int i=0; ca->caids[i]; i++) printf(" %04x",ca->caids[i]);
        printf("\n");
        }
      }
#endif
    caidTimer.Set(CAID_TIME);
    }
}

char *cSoftCAM::CurrKeyStr(int CardNum, int num)
{
  cScDvbDevice *dev=dynamic_cast<cScDvbDevice *>(cDevice::GetDevice(CardNum));
  char *str=0;
  if(dev && dev->Cam()) str=dev->Cam()->CurrentKeyStr(num);
  if(!str && num==0 && CapCheck(CardNum)) str=strdup(tr("(none)"));
  return str;
}

bool cSoftCAM::Active(void)
{
  for(int n=cDevice::NumDevices(); --n>=0;) {
    cScDvbDevice *dev=dynamic_cast<cScDvbDevice *>(cDevice::GetDevice(n));
    if(dev && dev->Cam() && dev->Cam()->Active()) return true;
    }
  return false;
}

void cSoftCAM::HouseKeeping(void)
{
  for(int n=cDevice::NumDevices(); --n>=0;) {
    cScDvbDevice *dev=dynamic_cast<cScDvbDevice *>(cDevice::GetDevice(n));
    if(dev && dev->Cam()) dev->Cam()->HouseKeeping();
    }
  if(Feature.KeyFile()) keys.HouseKeeping();
  PredictCaids(false);
}

bool cSoftCAM::Ignore(unsigned short caid)
{
  for(int i=0; i<MAXCAIGN; i++)
    if(ScSetup.CaIgnore[i]==caid) return true;
  return false;
}

void cSoftCAM::SetLogStatus(int CardNum, const cEcmInfo *ecm, bool on)
{
  cScDvbDevice *dev=dynamic_cast<cScDvbDevice *>(cDevice::GetDevice(CardNum));
  if(dev && dev->Cam()) dev->Cam()->LogEcmStatus(ecm,on);
}

void cSoftCAM::SetForceBudget(int n)
{
   if(n>=0 && n<MAXDVBDEVICES) budget|=(1<<n);
}

bool cSoftCAM::ForceBudget(int n)
{
   return budget && (budget&(1<<n));
}

#ifndef STATICBUILD

// --- cScDll ------------------------------------------------------------------

class cScDll : public cSimpleItem {
private:
  char *fileName;
  void *handle;
public:
  cScDll(const char *FileName);
  ~cScDll();
  bool Load(void);
  };

cScDll::cScDll(const char *FileName)
{
  fileName=strdup(FileName);
  handle=0;
}

cScDll::~cScDll()
{
//  if(handle) dlclose(handle); //XXX causes segfault :(
  free(fileName);
}

bool cScDll::Load(void)
{
  char *base=rindex(fileName,'/');
  if(!base) base=fileName;
  dyn(printf("scdll: loading library: %s\n",base))
  if(!handle) {
    handle=dlopen(fileName,RTLD_NOW|RTLD_LOCAL);
    if(handle) return true;
    LogError("dload: %s: %s",base,dlerror());
    }
  return false;
}

// --- cScDlls -----------------------------------------------------------------

#define LIBSC_PREFIX  "libsc-"
#define SO_INDICATOR   ".so."

class cScDlls : public cSimpleList<cScDll> {
private:
  void *handle;
public:
  cScDlls(void);
  ~cScDlls();
  bool Load(void);
  };

cScDlls::cScDlls(void)
{
  handle=0;
}

cScDlls::~cScDlls()
{
  Clear();
  if(handle) dlclose(handle);
  dyn(printf("scdlls: unload done\n"))
}

bool cScDlls::Load(void)
{
  Dl_info info;
  static int marker=0;
  if(!dladdr((void *)&marker,&info)) {
    LogError("dladdr: %s",dlerror());
    return false;
    }

  // we have to re-dlopen our selfs as VDR doesn't use RTLD_GLOBAL
  // but our symbols have to be available to the sub libs.
  handle=dlopen(info.dli_fname,RTLD_NOW|RTLD_GLOBAL);
  if(!handle) {
    LogError("dlopen myself: %s",dlerror());
    return false;
    }

  char *p;
  if((p=rindex(info.dli_fname,'/'))) *p=0;
  dyn(printf("scdlls: library path %s\n",info.dli_fname))

  char pat[32];
  snprintf(pat,sizeof(pat),"%s*-%d%s%s",LIBSC_PREFIX,SCAPIVERS,SO_INDICATOR,APIVERSION);
  bool res=true;
  cReadDir dir(info.dli_fname);
  struct dirent *e;
  while((e=dir.Next())) {
    if(!fnmatch(pat,e->d_name,FNM_PATHNAME|FNM_NOESCAPE)) {
      cScDll *dll=new cScDll(AddDirectory(info.dli_fname,e->d_name));
      if(dll) {
        if(!dll->Load()) res=false;
        Ins(dll);
        }
      }
    }
  return res;
}

#endif

// --- cScPlugin ---------------------------------------------------------------

class cScPlugin : public cPlugin {
private:
  bool setup;
#ifndef STATICBUILD
  cScDlls dlls;
#endif
public:
  cScPlugin(void);
  virtual ~cScPlugin();
  virtual const char *Version(void);
  virtual const char *Description(void);
  virtual const char *CommandLineHelp(void);
  virtual bool ProcessArgs(int argc, char *argv[]);
  virtual bool Start(void);
  virtual void Stop(void);
  virtual void Housekeeping(void);
  virtual cMenuSetupPage *SetupMenu(void);
  virtual bool SetupParse(const char *Name, const char *Value);
  virtual const char **SVDRPHelpPages(void);
  virtual cString SVDRPCommand(const char *Command, const char *Option, int &ReplyCode);
  };

cScPlugin::cScPlugin(void)
{
  static const char *logg[] = { "off","active CAIDs","all CAIDs" };
  ScOpts=new cOpts(0,5);
  ScOpts->Add(new cOptSel  ("AutoUpdate"   ,"Update keys (AU)"     ,&ScSetup.AutoUpdate,3,logg));
  ScOpts->Add(new cOptBool ("ConcurrentFF" ,"Concurrent FF streams",&ScSetup.ConcurrentFF));
  ScOpts->Add(new cOptBool ("LocalPriority","Prefer local systems" ,&ScSetup.LocalPriority));
  ScOpts->Add(new cOptMInt ("ScCaps"       ,"Active on DVB card"   , ScSetup.ScCaps,MAXSCCAPS,0));
  ScOpts->Add(new cOptMInt ("CaIgnore"     ,"Ignore CAID"          , ScSetup.CaIgnore,MAXCAIGN,2));
  setup=false;
#ifndef STATICBUILD
  dlls.Load();
#endif
  cScDvbDevice::Capture();
}

cScPlugin::~cScPlugin()
{
  delete ScOpts;
}

bool cScPlugin::Start(void)
{
  d(printf("scplugin: version %s starting\n",VERSION))
  if(APIVERSNUM<MINAPIVERSNUM) {
    LogError("SC plugin needs at least VDR API version %d.%d.%d",MIN_VERS,MIN_MAJOR,MIN_MINOR);
    return false;
    }
  if(sizeof(int)!=4) {
    LogError("compiled with 'int' as %d bit. Only supporting 32 bit.",(int)sizeof(int)*8);
    return false;
    }
  if(sizeof(long long)!=8) {
    LogError("compiled with 'long long' as %d bit. Only supporting 64 bit.",(int)sizeof(long long)*8);
    return false;
    }
    
  ScPlugin=this;
  Feature.AddPhrases(ScPhrases);
  RegisterI18n(Feature.GetPhrases());
  filemaps.SetCfgDir(ConfigDirectory());
  ScSetup.CheckConfig();
  if(!cSoftCAM::Load(ConfigDirectory())) return false;
  if(Feature.SmartCard()) {
#ifdef DEFAULT_PORT
    smartcards.AddPort(DEFAULT_PORT);
#endif
    smartcards.LaunchWatcher();
    }
  setup=true;
  cScDvbDevice::Startup();
  return true;
}

void cScPlugin::Stop(void)
{
  setup=false;
  cSoftCAM::ShutDown();
  cScDvbDevice::Shutdown();
  d(printf("scplugin: cleanup done\n"))
}

const char *cScPlugin::Version(void)
{
  return VERSION;
}

const char *cScPlugin::Description(void)
{
  return tr(DESCRIPTION);
}

const char *cScPlugin::CommandLineHelp(void)
{
  static char *help_str=0;
  
  free(help_str);    //                                     for easier orientation, this is column 80|
  asprintf(&help_str,"  -B N      --budget=N     forces DVB device N to budget mode (using FFdecsa)\n"
                     "  -I        --inverse-cd   use inverse CD detection for the next serial device\n"
                     "  -R        --inverse-rst  use inverse RESET for the next serial device\n"
                     "  -C FREQ   --clock=FREQ   use FREQ as clock for the card reader on the next\n"
                     "                           serial device (rather than 3.5712 MHz\n"
                     "  -s DEV,   --serial=DEV   activate Phoenix ISO interface on serial device DEV\n"
                     "                           (default: %s)\n"
                     "  -d CMD,   --dialup=CMD   call CMD to start/stop dialup-network\n"
                     "                           (default: %s)\n"
                     "  -t SECS,  --timeout=SECS shutdown timeout for dialup-network\n"
                     "                           (default: %d secs)\n",
                     "none","none",netTimeout/1000
                     );
  return help_str;
}

bool cScPlugin::ProcessArgs(int argc, char *argv[])
{
  static struct option long_options[] = {
      { "serial",      required_argument, NULL, 's' },
      { "inverse-cd",  no_argument,       NULL, 'I' },
      { "inverse-rst", no_argument,       NULL, 'R' },
      { "clock",       required_argument, NULL, 'C' },
      { "dialup",      required_argument, NULL, 'd' },
      { "external-au", required_argument, NULL, 'E' },
      { "budget",      required_argument, NULL, 'B' },
      { NULL }
    };

  int c, option_index=0;
  bool invCD=false, invRST=false;
  int clock=0;
  while((c=getopt_long(argc,argv,"d:s:t:B:C:E:IR",long_options,&option_index))!=-1) {
    switch (c) {
      case 'I': invCD=true; break;
      case 'R': invRST=true; break;
      case 'C': clock=atoi(optarg); break;
      case 's': smartcards.AddPort(optarg,invCD,invRST,clock); invCD=false; invRST=false; clock=0; break;
      case 'd': netscript=optarg; break;
      case 't': netTimeout=atoi(optarg)*1000; break;
      case 'E': externalAU=optarg; break;
      case 'B': cSoftCAM::SetForceBudget(atoi(optarg)); break;
      default:  return false;
      }
    }
  return true;
}

cMenuSetupPage *cScPlugin::SetupMenu(void)
{
  return new cMenuSetupSc(ConfigDirectory());
}

bool cScPlugin::SetupParse(const char *Name, const char *Value)
{
  if(!ScOpts->Parse(Name,Value) && !cSystems::ConfigParse(Name,Value))
    return false;
  return true;
}

void cScPlugin::Housekeeping(void)
{
  cSoftCAM::HouseKeeping();
}

const char **cScPlugin::SVDRPHelpPages(void)
{
  static const char *HelpPages[] = {
    "RELOAD\n"
    "    Reload all configuration files.",
    NULL
    };
  return HelpPages;
}

cString cScPlugin::SVDRPCommand(const char *Command, const char *Option, int &ReplyCode)
{
  if(!strcasecmp(Command,"RELOAD")) {
    if(cSoftCAM::Active()) {
      ReplyCode=550;
      return "Softcam active. Can't reload files now";
      }
    else {
      if(cSoftCAM::Load(ConfigDirectory()))
        return "Files reloaded successfully";
      else {
        ReplyCode=901;
        return "Reloading files not entirely successfull";
        }
      }
    }
  return NULL;
}

VDRPLUGINCREATOR(cScPlugin); // Don't touch this!
